# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-21 08:55
from __future__ import unicode_literals

import uuid

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
import localized_fields.fields.field
import localized_fields.fields.text_field
import simple_history.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0019_remove_answer_value_document")]

    operations = [
        migrations.CreateModel(
            name="HistoricalAnswer",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "value",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ("date", models.DateField(blank=True, null=True)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Document",
                    ),
                ),
                (
                    "file",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.File",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical answer",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalAnswerDocument",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "answer",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Answer",
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Document",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical answer document",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalDocument",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "family",
                    models.UUIDField(
                        db_index=True, help_text="Family id which document belongs too."
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "form",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical document",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalFile",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical file",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalForm",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("slug", models.SlugField()),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ("is_published", models.BooleanField(default=False)),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Reference this form has been copied from",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical form",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalFormQuestion",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "form",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical form question",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalOption",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("slug", models.SlugField()),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Reference this option has been copied from",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Option",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical option",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalQuestion",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("slug", models.SlugField()),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("multiple_choice", "multiple_choice"),
                            ("integer", "integer"),
                            ("float", "float"),
                            ("date", "date"),
                            ("choice", "choice"),
                            ("textarea", "textarea"),
                            ("text", "text"),
                            ("table", "table"),
                            ("form", "form"),
                            ("file", "file"),
                            ("dynamic_choice", "dynamic_choice"),
                            ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                            ("static", "static"),
                        ],
                        max_length=23,
                    ),
                ),
                ("is_required", models.TextField(default="false")),
                ("is_hidden", models.TextField(default="false")),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "placeholder",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                (
                    "info_text",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                (
                    "static_content",
                    localized_fields.fields.text_field.LocalizedTextField(
                        blank=True, null=True, required=[]
                    ),
                ),
                (
                    "configuration",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "data_source",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "format_validators",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        blank=True,
                        default=list,
                        size=None,
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "row_form",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Form that represents rows of a TableQuestion",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Reference this question has been copied from",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Question",
                    ),
                ),
                (
                    "sub_form",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Form referenced in a FormQuestion",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical question",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalQuestionOption",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("id", models.CharField(db_index=True, max_length=255)),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "option",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Option",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical question option",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



import pytest

from caluma.caluma_workflow.dynamic_groups import (
    BaseDynamicGroups,
    register_dynamic_group,
)
from caluma.caluma_workflow.jexl import GroupJexl


class CustomDynamicGroups(BaseDynamicGroups):
    @register_dynamic_group("admin")
    def resolve_admin(self, task, case, user, prev_work_item):
        return "admin-group-slug"

    @register_dynamic_group("staff")
    def resolve_staff(self, task, case, user, prev_work_item):
        return ["staff-a", "staff-b"]


@pytest.mark.parametrize("case__created_by_group", ["group1"])
@pytest.mark.parametrize(
    "expression,expected_result",
    [
        ('["admin"]|groups', ["admin-group-slug"]),
        ('["staff"]|groups', ["staff-a", "staff-b"]),
        ('["admin", "staff"]|groups', ["admin-group-slug", "staff-a", "staff-b"]),
        ('["some-other-group"]|groups', ["some-other-group"]),
    ],
)
def test_dynamic_groups(
    db, task, case, admin_user, work_item, expression, expected_result
):
    GroupJexl.dynamic_groups_classes = [CustomDynamicGroups]

    jexl = GroupJexl(None, task, case, admin_user, work_item)

    assert sorted(jexl.evaluate(expression)) == sorted(expected_result)



import pytest

from ..jexl import FlowJexl, GroupJexl


@pytest.mark.parametrize(
    "expression,expected_tasks",
    [
        (
            "63 > 62 ? 'test'|task : ['test2', 'test3']|tasks",
            {"test", "test2", "test3"},
        ),
        ("['test2', 'test3']|tasks", {"test2", "test3"}),
        ("'test'|task", {"test"}),
    ],
)
def test_flow_extract_tasks(expression, expected_tasks):
    jexl = FlowJexl()

    assert expected_tasks == set(jexl.extract_tasks(expression))


@pytest.mark.parametrize(
    "expression,num_errors",
    [
        # correct cases
        ('"task-slug"|task', 0),
        ('["task-slug1", "task-slug2"]|tasks', 0),
        # invalid subject type
        ("100|task", 1),
        ("'task-slug'|tasks", 1),
    ],
)
def test_flow_jexl_validate(expression, num_errors):
    jexl = FlowJexl()
    assert len(list(jexl.validate(expression))) == num_errors


@pytest.mark.parametrize(
    "expression,num_errors",
    [
        # correct cases
        ("['group1']|groups", 0),
        # invalid subject type
        ("100|groups", 1),
        ("'task-slug'|groups", 1),
    ],
)
def test_group_jexl_validate(expression, num_errors):
    jexl = GroupJexl()
    assert len(list(jexl.validate(expression))) == num_errors



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-03-19 17:20
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0007_auto_20190208_1232")]

    operations = [
        migrations.AlterField(
            model_name="answer",
            name="value",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                ],
                max_length=15,
            ),
        ),
    ]



import pytest

from ...caluma_core.tests import (
    extract_global_id_input_fields,
    extract_serializer_input_fields,
)
from .. import models, serializers


@pytest.mark.parametrize(
    "question__type,question__configuration,question__data_source,question__format_validators",
    [
        (models.Question.TYPE_INTEGER, {"max_value": 10, "min_value": 0}, None, []),
        (models.Question.TYPE_FLOAT, {"max_value": 1.0, "min_value": 0.0}, None, []),
        (models.Question.TYPE_FLOAT, {}, None, []),
        (models.Question.TYPE_DATE, {}, None, []),
        (models.Question.TYPE_TEXT, {"min_length": 10}, None, ["email"]),
        (models.Question.TYPE_TEXTAREA, {"max_length": 10}, None, []),
        (models.Question.TYPE_CHOICE, {}, None, []),
        (models.Question.TYPE_MULTIPLE_CHOICE, {}, None, []),
        (models.Question.TYPE_FORM, {}, None, []),
        (models.Question.TYPE_FILE, {}, None, []),
        (models.Question.TYPE_DYNAMIC_CHOICE, {}, "MyDataSource", []),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, {}, "MyDataSource", []),
        (models.Question.TYPE_STATIC, {}, None, []),
    ],
)
def test_query_all_questions(
    schema_executor,
    db,
    snapshot,
    question,
    form,
    form_question_factory,
    question_option,
    data_source_settings,
):
    form_question_factory.create(form=form)

    query = """
        query AllQuestionsQuery($search: String, $forms: [ID]) {
          allQuestions(isArchived: false, search: $search, excludeForms: $forms) {
            totalCount
            edges {
              node {
                id
                __typename
                slug
                label
                meta
                infoText
                ... on TextQuestion {
                  minLength
                  maxLength
                  placeholder
                  formatValidators {
                    edges {
                      node {
                        slug
                        name
                        regex
                        errorMsg
                      }
                    }
                  }
                }
                ... on TextareaQuestion {
                  minLength
                  maxLength
                  placeholder
                  formatValidators {
                    edges {
                      node {
                        slug
                        name
                        regex
                        errorMsg
                      }
                    }
                  }
                }
                ... on FloatQuestion {
                  floatMinValue: minValue
                  floatMaxValue: maxValue
                  placeholder
                }
                ... on IntegerQuestion {
                  integerMinValue: minValue
                  integerMaxValue: maxValue
                  placeholder
                }
                ... on MultipleChoiceQuestion {
                  options {
                    totalCount
                    edges {
                      node {
                        slug
                      }
                    }
                  }
                }
                ... on ChoiceQuestion {
                  options {
                    totalCount
                    edges {
                      node {
                        slug
                      }
                    }
                  }
                }
                ... on DynamicMultipleChoiceQuestion {
                  options {
                    edges {
                      node {
                        slug
                        label
                      }
                    }
                  }
                }
                ... on DynamicChoiceQuestion {
                  options {
                    edges {
                      node {
                        slug
                        label
                      }
                    }
                  }
                }
                ... on FormQuestion {
                  subForm {
                    slug
                  }
                }
                ... on StaticQuestion {
                  staticContent
                }
              }
            }
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "search": question.label,
            "forms": [extract_global_id_input_fields(form)["id"]],
        },
    )

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__meta", [{"meta": "set"}])
def test_copy_question(
    db, snapshot, question, question_option_factory, schema_executor
):
    question_option_factory.create_batch(5, question=question)
    query = """
        mutation CopyQuestion($input: CopyQuestionInput!) {
          copyQuestion(input: $input) {
            question {
              slug
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": {
            "source": question.pk,
            "slug": "new-question",
            "label": "Test Question",
        }
    }
    result = schema_executor(query, variable_values=inp)

    assert not result.errors

    question_slug = result.data["copyQuestion"]["question"]["slug"]
    assert question_slug == "new-question"
    new_question = models.Question.objects.get(pk=question_slug)
    assert new_question.label == "Test Question"
    assert new_question.meta == question.meta
    assert new_question.type == question.type
    assert new_question.configuration == question.configuration
    assert new_question.row_form == question.row_form
    assert new_question.is_hidden == question.is_hidden
    assert new_question.is_required == question.is_required
    assert new_question.source == question
    assert list(
        models.QuestionOption.objects.filter(question=new_question).values("option")
    ) == list(models.QuestionOption.objects.filter(question=question).values("option"))


@pytest.mark.parametrize(
    "mutation",
    [
        "SaveTextQuestion",
        "SaveTextareaQuestion",
        "SaveIntegerQuestion",
        "SaveFloatQuestion",
        "SaveDateQuestion",
        "SaveFileQuestion",
    ],
)
@pytest.mark.parametrize(
    "question__is_required,success", (("true", True), ("true|invalid", False))
)
def test_save_question(db, snapshot, question, mutation, schema_executor, success):
    mutation_func = mutation[0].lower() + mutation[1:]
    query = f"""
        mutation {mutation}($input: {mutation}Input!) {{
          {mutation_func}(input: $input) {{
            question {{
              id
              slug
              label
              meta
              __typename
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "question__type,question__configuration,question__format_validators,success",
    [
        (models.Question.TYPE_TEXT, {"max_length": 10}, [], True),
        (models.Question.TYPE_TEXT, {"max_length": 10}, ["email"], True),
        (models.Question.TYPE_TEXT, {"max_length": 10}, ["notavalidator"], False),
    ],
)
def test_save_text_question(db, question, schema_executor, success):
    query = """
        mutation SaveTextQuestion($input: SaveTextQuestionInput!) {
          saveTextQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TextQuestion {
                maxLength
                formatValidators {
                  edges {
                    node {
                      slug
                      name
                      regex
                      errorMsg
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTextQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        assert result.data["saveTextQuestion"]["question"]["maxLength"] == 10
        if question.format_validators:
            assert (
                result.data["saveTextQuestion"]["question"]["formatValidators"][
                    "edges"
                ][0]["node"]["slug"]
                == "email"
            )


@pytest.mark.parametrize(
    "question__type,question__configuration",
    [(models.Question.TYPE_TEXTAREA, {"max_length": 10})],
)
def test_save_textarea_question(db, question, schema_executor):
    query = """
        mutation SaveTextareaQuestion($input: SaveTextareaQuestionInput!) {
          saveTextareaQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TextareaQuestion {
                maxLength
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTextareaQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    assert result.data["saveTextareaQuestion"]["question"]["maxLength"] == 10


@pytest.mark.parametrize(
    "question__type,question__configuration,success",
    [
        (models.Question.TYPE_FLOAT, {"max_value": 10.0, "min_value": 0.0}, True),
        (models.Question.TYPE_FLOAT, {"max_value": 1.0, "min_value": 10.0}, False),
    ],
)
def test_save_float_question(db, snapshot, question, schema_executor, success):
    query = """
        mutation SaveFloatQuestion($input: SaveFloatQuestionInput!) {
          saveFloatQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on FloatQuestion {
                minValue
                maxValue
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveFloatQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "question__type,question__configuration,success",
    [
        (models.Question.TYPE_INTEGER, {"max_value": 10, "min_value": 0}, True),
        (models.Question.TYPE_INTEGER, {"max_value": 1, "min_value": 10}, False),
    ],
)
def test_save_integer_question(db, snapshot, question, success, schema_executor):
    query = """
        mutation SaveIntegerQuestion($input: SaveIntegerQuestionInput!) {
          saveIntegerQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on FloatQuestion {
                minValue
                maxValue
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveIntegerQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_MULTIPLE_CHOICE])
def test_save_multiple_choice_question(
    db, snapshot, question, question_option_factory, schema_executor
):
    question_option_factory.create_batch(2, question=question)

    option_ids = question.options.order_by("-slug").values_list("slug", flat=True)

    query = """
        mutation SaveMultipleChoiceQuestion($input: SaveMultipleChoiceQuestionInput!) {
          saveMultipleChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on MultipleChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveMultipleChoiceQuestionSerializer, question
        )
    }
    inp["input"]["options"] = option_ids
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_CHOICE])
def test_save_choice_question(db, snapshot, question, question_option, schema_executor):
    query = """
        mutation SaveChoiceQuestion($input: SaveChoiceQuestionInput!) {
          saveChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on ChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveChoiceQuestionSerializer, question
        )
    }
    question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete", [True, False])
@pytest.mark.parametrize(
    "question__type",
    [models.Question.TYPE_DYNAMIC_CHOICE, models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE],
)
def test_save_dynamic_choice_question(
    db, snapshot, question, delete, schema_executor, data_source_settings
):
    query = """
        mutation SaveDynamicChoiceQuestion($input: SaveDynamicChoiceQuestionInput!) {
          saveDynamicChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on DynamicChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveDynamicChoiceQuestionSerializer, question
        )
    }
    if delete:
        question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete", [True, False])
@pytest.mark.parametrize(
    "question__type", [models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE]
)
def test_save_dynamic_multiple_choice_question(
    db,
    snapshot,
    question,
    delete,
    question_option_factory,
    schema_executor,
    data_source_settings,
):
    query = """
        mutation SaveDynamicMultipleChoiceQuestion($input: SaveDynamicMultipleChoiceQuestionInput!) {
          saveDynamicMultipleChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on DynamicMultipleChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveDynamicMultipleChoiceQuestionSerializer, question
        )
    }
    if delete:
        question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_TABLE])
def test_save_table_question(db, snapshot, question, question_option, schema_executor):
    query = """
        mutation SaveTableQuestion($input: SaveTableQuestionInput!) {
          saveTableQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TableQuestion {
                rowForm {
                  slug
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTableQuestionSerializer, question
        )
    }
    question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_FORM])
def test_save_form_question(db, snapshot, question, question_option, schema_executor):
    query = """
        mutation SaveFormQuestion($input: SaveFormQuestionInput!) {
          saveFormQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on FormQuestion {
                subForm {
                  slug
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveFormQuestionSerializer, question
        )
    }
    question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_STATIC])
def test_save_static_question(db, snapshot, question, schema_executor):
    query = """
        mutation SaveStaticQuestion($input: SaveStaticQuestionInput!) {
          saveStaticQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on StaticQuestion {
                staticContent
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveStaticQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors)
    snapshot.assert_match(result.data)


def test_all_questions_slug_filter(
    db,
    schema_executor,
    question_option_factory,
    question_factory,
    form_factory,
    form_question_factory,
    option_factory,
):
    query = """
        query FilteredQuestions {
          allQuestions {
            edges {
              node {
                slug
              }
            }
          }
        }
    """
    questions = question_factory.create_batch(4)

    forms = form_factory.create_batch(2)

    for question in questions:
        for form in forms:
            question.forms.through.objects.create(form=form, question=question)

    num_questions = models.Question.objects.all().count()
    result = schema_executor(query)

    assert not result.errors
    assert len(result.data["allQuestions"]["edges"]) == num_questions



from django.conf import settings
from django.conf.urls import url

from caluma.caluma_user.views import AuthenticationGraphQLView

urlpatterns = [
    url(r"", AuthenticationGraphQLView.as_view(graphiql=settings.DEBUG), name="graphql")
]



import functools

import pytest
from pyjexl import JEXL

from ..jexl import Cache, ExtractTransformSubjectAnalyzer


@pytest.mark.parametrize(
    "expression, expected_transforms",
    [
        ('63 > 62 ? "test"|transform1 : "test2"|transform2', {"test"}),
        ('63 > 62 ? "test"|transform1 : "test2"|transform1', {"test", "test2"}),
        ('"test2"|transform2', set()),
        ('"test2"|transform1', {"test2"}),
    ],
)
def test_extract_transforms(expression, expected_transforms):
    jexl = JEXL()
    jexl.add_transform("transform1", lambda x: x)
    jexl.add_transform("transform2", lambda x: x)

    assert set(
        jexl.analyze(
            expression,
            functools.partial(
                ExtractTransformSubjectAnalyzer, transforms=["transform1"]
            ),
        )
    ) == set(expected_transforms)


def test_jexl_cache():
    cache = Cache(20, 10)

    # fill the cache "to the brim"
    for x in range(19):
        cache.get_or_set(x, lambda: x)
    assert len(cache._mru) == 19
    assert len(cache._cache) == 19

    # insert last element before eviction
    cache.get_or_set("x", lambda: "x")
    assert len(cache._mru) == 20
    assert len(cache._cache) == 20

    # one more - this should trigger eviction
    cache.get_or_set("y", lambda: "y")
    assert len(cache._mru) == 10
    assert len(cache._cache) == 10

    # now make sure the right entries were evicted
    for x in range(10):
        assert x not in cache._cache
        assert x not in cache._mru

    # validate invariants
    assert cache._cache.keys() == cache._mru.keys()



import json

import pytest
from graphene.utils.str_converters import to_const

from ...caluma_core.relay import extract_global_id
from ...caluma_form.models import Question
from ...caluma_user.models import BaseUser
from .. import api, models


def test_query_all_work_items_filter_status(db, work_item_factory, schema_executor):
    work_item_factory(status=models.WorkItem.STATUS_READY)
    work_item_factory(status=models.WorkItem.STATUS_COMPLETED)

    query = """
        query WorkItems($status: WorkItemStatusArgument!) {
          allWorkItems(status: $status) {
            totalCount
            edges {
              node {
                status
              }
            }
          }
        }
    """

    result = schema_executor(
        query, variable_values={"status": to_const(models.WorkItem.STATUS_READY)}
    )

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == 1
    assert result.data["allWorkItems"]["edges"][0]["node"]["status"] == to_const(
        models.WorkItem.STATUS_READY
    )


@pytest.mark.parametrize("key", ["addressed_groups", "controlling_groups"])
def test_query_all_work_items_filter_groups(
    db, key, work_item_factory, schema_executor
):
    factory_kwargs = {key: ["A", "B"]}
    work_item_factory(**factory_kwargs)

    query = """
        query WorkItems($groups: [String]!) {
          allWorkItems(addressedGroups: $groups) {
            edges {
              node {
                addressedGroups
              }
            }
          }
        }
    """

    if key == "controlling_groups":
        query = """
            query WorkItems($groups: [String]!) {
              allWorkItems(controllingGroups: $groups) {
                edges {
                  node {
                    controllingGroups
                  }
                }
              }
            }
        """

    result = schema_executor(query, variable_values={"groups": ["B", "C"]})

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == 1
    assert result.data["allWorkItems"]["edges"][0]["node"][key.replace("_g", "G")] == [
        "A",
        "B",
    ]

    result = schema_executor(query, variable_values={"groups": ["C", "D"]})

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == 0


@pytest.mark.parametrize("task__type,task__form", [(models.Task.TYPE_SIMPLE, None)])
@pytest.mark.parametrize(
    "work_item__status,case__status,success",
    [
        (models.WorkItem.STATUS_READY, models.Case.STATUS_COMPLETED, True),
        (models.WorkItem.STATUS_COMPLETED, models.Case.STATUS_COMPLETED, False),
        (models.WorkItem.STATUS_READY, models.Case.STATUS_RUNNING, False),
    ],
)
def test_complete_work_item_last(
    db, snapshot, work_item, success, admin_schema_executor
):
    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              closedByUser
              status
              case {
                closedByUser
                status

              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = admin_schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,case__status,task__type,question__type,answer__value,success",
    [
        (
            models.WorkItem.STATUS_READY,
            None,
            models.Case.STATUS_RUNNING,
            models.Task.TYPE_COMPLETE_WORKFLOW_FORM,
            Question.TYPE_FLOAT,
            1.0,
            True,
        ),
        (
            models.WorkItem.STATUS_READY,
            None,
            models.Case.STATUS_RUNNING,
            models.Task.TYPE_COMPLETE_WORKFLOW_FORM,
            Question.TYPE_TEXT,
            "",
            False,
        ),
    ],
)
def test_complete_workflow_form_work_item(
    db,
    work_item,
    answer,
    question_factory,
    answer_factory,
    answer_document_factory,
    form_question,
    success,
    schema_executor,
):
    table_question = question_factory(type=Question.TYPE_TABLE)
    table_answer = answer_factory(
        question=table_question, document=answer.document, value=None
    )
    answer_document = answer_document_factory(answer=table_answer)
    answer_document.document.answers.add(answer_factory())

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        assert result.data["completeWorkItem"]["workItem"]["status"] == to_const(
            models.WorkItem.STATUS_COMPLETED
        )
        assert result.data["completeWorkItem"]["workItem"]["case"][
            "status"
        ] == to_const(models.Case.STATUS_COMPLETED)


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,case__status,task__type,case__document",
    [
        (
            models.WorkItem.STATUS_READY,
            None,
            models.Case.STATUS_RUNNING,
            models.Task.TYPE_COMPLETE_TASK_FORM,
            None,
        )
    ],
)
@pytest.mark.parametrize(
    "question__type,answer__value,success",
    [(Question.TYPE_INTEGER, 1, True), (Question.TYPE_CHOICE, "", False)],
)
def test_complete_task_form_work_item(
    db, work_item, answer, form_question, success, schema_executor
):
    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        assert result.data["completeWorkItem"]["workItem"]["status"] == to_const(
            models.WorkItem.STATUS_COMPLETED
        )
        assert result.data["completeWorkItem"]["workItem"]["case"][
            "status"
        ] == to_const(models.Case.STATUS_COMPLETED)


@pytest.mark.parametrize("question__type,answer__value", [(Question.TYPE_INTEGER, 1)])
def test_complete_multiple_instance_task_form_work_item(
    db, task_factory, work_item_factory, answer, form_question, schema_executor
):
    task = task_factory(is_multiple_instance=True)
    work_item_1 = work_item_factory(task=task, child_case=None)
    work_item_2 = work_item_factory(task=task, child_case=None, case=work_item_1.case)
    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item_1.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors)
    assert result.data["completeWorkItem"]["workItem"]["status"] == to_const(
        models.WorkItem.STATUS_COMPLETED
    )
    assert result.data["completeWorkItem"]["workItem"]["case"]["status"] == to_const(
        models.Case.STATUS_RUNNING
    )

    inp = {"input": {"id": work_item_2.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors)
    assert result.data["completeWorkItem"]["workItem"]["status"] == to_const(
        models.WorkItem.STATUS_COMPLETED
    )
    assert result.data["completeWorkItem"]["workItem"]["case"]["status"] == to_const(
        models.Case.STATUS_COMPLETED
    )


@pytest.mark.parametrize("question__type,answer__value", [(Question.TYPE_INTEGER, 1)])
def test_complete_multiple_instance_task_form_work_item_next(
    db,
    task_factory,
    task_flow_factory,
    work_item_factory,
    answer,
    form_question,
    snapshot,
    schema_executor,
):
    task = task_factory(is_multiple_instance=True)
    work_item = work_item_factory(task=task, child_case=None)
    work_item_factory(
        task=task,
        child_case=None,
        status=models.WorkItem.STATUS_COMPLETED,
        case=work_item.case,
    )

    task_next = task_factory(
        type=models.Task.TYPE_SIMPLE, form=None, address_groups='["group-name"]|groups'
    )
    task_flow = task_flow_factory(task=task, workflow=work_item.case.workflow)
    task_flow.flow.next = f"'{task_next.slug}'|task"
    task_flow.flow.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
                workItems(orderBy: STATUS_DESC) {
                  totalCount
                  edges {
                    node {
                      status
                      addressedGroups
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors)
    snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,task__type,work_item__controlling_groups,work_item__created_by_group,case__created_by_group",
    [
        (
            models.WorkItem.STATUS_READY,
            None,
            models.Task.TYPE_SIMPLE,
            ["controlling-group1", "controlling-group2"],
            "work-item-creating-group",
            "case-creating-group",
        )
    ],
)
@pytest.mark.parametrize(
    "group_jexl",
    [
        '["some-group"]|groups',
        "info.prev_work_item.controlling_groups",
        "info.work_item.created_by_group",
        "info.case.created_by_group",
        "[info.case.created_by_group, info.work_item.created_by_group]|groups",
        "[info.case.created_by_group, info.work_item.created_by_group]",
    ],
)
def test_complete_work_item_with_next(
    db,
    group_jexl,
    snapshot,
    work_item,
    task,
    task_factory,
    task_flow_factory,
    workflow,
    info,
    case,
    schema_executor,
):
    class FakeUser(BaseUser):
        def __init__(self):
            self.username = "foo"
            self.groups = ["fake-user-group"]

        @property
        def group(self):
            return self.groups[0]

    info.context.user = FakeUser()

    task_next = task_factory(
        type=models.Task.TYPE_SIMPLE,
        form=None,
        address_groups=group_jexl,
        control_groups=group_jexl,
    )
    task_flow = task_flow_factory(task=task, workflow=workflow)
    task_flow.flow.next = f"'{task_next.slug}'|task"
    task_flow.flow.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
                workItems(orderBy: STATUS_DESC) {
                  totalCount
                  edges {
                    node {
                      status
                      addressedGroups
                      controllingGroups
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp, info=info)

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,task__type",
    [(models.WorkItem.STATUS_READY, None, models.Task.TYPE_SIMPLE)],
)
def test_complete_work_item_with_next_multiple_tasks(
    db,
    case,
    work_item,
    task,
    task_factory,
    task_flow_factory,
    workflow,
    schema_executor,
):
    task_next_1, task_next_2 = task_factory.create_batch(
        2, type=models.Task.TYPE_SIMPLE
    )
    task_flow = task_flow_factory(task=task, workflow=workflow)
    task_flow.flow.next = f"['{task_next_1.slug}', '{task_next_2.slug}']|task"
    task_flow.flow.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
                workItems {
                  totalCount
                  edges {
                    node {
                      status
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    assert case.work_items.count() == 3
    assert set(
        case.work_items.values_list("task", flat=True).filter(
            status=models.WorkItem.STATUS_READY
        )
    ) == {task_next_1.pk, task_next_2.pk}


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,task__type",
    [(models.WorkItem.STATUS_READY, None, models.Task.TYPE_SIMPLE)],
)
def test_complete_work_item_with_next_multiple_instance_task(
    db,
    case,
    work_item,
    task,
    task_factory,
    task_flow_factory,
    workflow,
    schema_executor,
):
    task_next = task_factory.create(
        is_multiple_instance=True, address_groups=["group1", "group2", "group3"]
    )
    task_flow = task_flow_factory(task=task, workflow=workflow)
    task_flow.flow.next = f"['{task_next.slug}']|task"
    task_flow.flow.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              status
              case {
                status
                workItems {
                  totalCount
                  edges {
                    node {
                      status
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    assert case.work_items.filter(status=models.WorkItem.STATUS_READY).count() == 3
    assert case.work_items.filter(status=models.WorkItem.STATUS_COMPLETED).count() == 1
    for work_item in case.work_items.filter(status=models.WorkItem.STATUS_READY):
        assert len(work_item.addressed_groups) == 1


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,task__type",
    [(models.WorkItem.STATUS_COMPLETED, None, models.Task.TYPE_SIMPLE)],
)
def test_complete_work_item_with_merge(
    db,
    case,
    work_item,
    work_item_factory,
    task,
    task_factory,
    flow_factory,
    task_flow_factory,
    workflow,
    schema_executor,
):
    # create two work items which can be processed in parallel
    task_1, task_2 = task_factory.create_batch(2, type=models.Task.TYPE_SIMPLE)
    work_item_1 = work_item_factory(
        task=task_1, status=models.WorkItem.STATUS_READY, child_case=None, case=case
    )
    work_item_2 = work_item_factory(
        task=task_2, status=models.WorkItem.STATUS_READY, child_case=None, case=case
    )
    ready_workitems = case.work_items.filter(status=models.WorkItem.STATUS_READY)
    assert ready_workitems.count() == 2

    # both work item's tasks reference the same merge task
    task_merge = task_factory(type=models.Task.TYPE_COMPLETE_TASK_FORM)
    flow = flow_factory(next=f"'{task_merge.slug}'|task")
    task_flow_factory(task=work_item_1.task, workflow=workflow, flow=flow)
    task_flow_factory(task=work_item_2.task, workflow=workflow, flow=flow)

    # complete one of the work item
    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            workItem {
              id
            }
          }
        }
    """
    inp = {"input": {"id": work_item_1.pk}}
    result = schema_executor(query, variable_values=inp)
    assert not result.errors

    # one parallel work item is left, no new one created as both preceding
    # work items need to be completed first
    assert ready_workitems.count() == 1
    assert ready_workitems.first().pk == work_item_2.pk

    # complete second work item
    inp = {"input": {"id": work_item_2.pk}}
    result = schema_executor(query, variable_values=inp)
    assert not result.errors

    # new work item is created of merge task
    assert ready_workitems.count() == 1
    ready_workitem = ready_workitems.first()
    assert ready_workitem.task == task_merge
    assert ready_workitem.document_id is not None


@pytest.mark.parametrize(
    "work_item_name,work_item_description,expected_name,expected_description",
    [
        (
            None,
            None,
            {"en": "Task name", "de": None, "fr": None},
            {"en": "Task description", "de": None, "fr": None},
        ),
        (
            "WorkItem name",
            "WorkItem description",
            {"en": "WorkItem name", "de": None, "fr": None},
            {"en": "WorkItem description", "de": None, "fr": None},
        ),
    ],
)
@pytest.mark.parametrize(
    "task__name,task__description", [("Task name", "Task description")]
)
def test_save_work_item(
    db,
    work_item,
    work_item_name,
    work_item_description,
    expected_name,
    expected_description,
    schema_executor,
):
    query = """
        mutation SaveWorkItem($input: SaveWorkItemInput!) {
          saveWorkItem(input: $input) {
            clientMutationId
          }
        }
    """

    assigned_users = ["user1", "user2"]
    inp = {
        "input": {
            "workItem": str(work_item.pk),
            "assignedUsers": assigned_users,
            "meta": json.dumps({"test": "test"}),
            "name": work_item_name,
            "description": work_item_description,
        }
    }
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    work_item.refresh_from_db()
    assert work_item.assigned_users == assigned_users
    assert work_item.meta == {"test": "test"}
    assert dict(work_item.name) == expected_name
    assert dict(work_item.description) == expected_description


@pytest.mark.parametrize(
    "task__is_multiple_instance,task__control_groups,expected_groups,work_item__status,set_groups,success",
    [
        # Failing
        (False, None, None, models.WorkItem.STATUS_READY, False, False),
        (True, None, None, models.WorkItem.STATUS_COMPLETED, False, False),
        # Successful
        (
            True,
            '["groups-transform"]|groups',
            ["groups-transform"],
            models.WorkItem.STATUS_READY,
            False,
            True,
        ),
        (
            True,
            "info.case.created_by_group",
            ["group-case"],
            models.WorkItem.STATUS_READY,
            False,
            True,
        ),
        (
            True,
            "info.work_item.created_by_group",
            ["group-work-item"],
            models.WorkItem.STATUS_READY,
            False,
            True,
        ),
        # Controlling jexl evaluates to None
        (True, "", [], models.WorkItem.STATUS_READY, False, True),
        (True, None, [], models.WorkItem.STATUS_READY, False, True),
        (
            True,
            "info.prev_work_item.controlling_groups",
            [],
            models.WorkItem.STATUS_READY,
            False,
            True,
        ),
        # Reset controlling_groups and addressed_groups
        (
            True,
            "info.work_item.created_by_group",
            [],
            models.WorkItem.STATUS_READY,
            True,
            True,
        ),
    ],
)
@pytest.mark.parametrize("case__created_by_group", ["group-case"])
@pytest.mark.parametrize("work_item__child_case", [None])
@pytest.mark.parametrize(
    "task__name,task__description", [("Task name", "Task description")]
)
def test_create_work_item(
    db, work_item, task, expected_groups, set_groups, success, schema_executor, info
):
    task.address_groups = task.control_groups
    task.save()

    class FakeUser(BaseUser):
        def __init__(self):
            self.username = "foo"
            self.groups = ["group-work-item"]

        @property
        def group(self):
            return self.groups[0]

    info.context.user = FakeUser()

    query = """
        mutation CreateWorkItem($input: CreateWorkItemInput!) {
          createWorkItem(input: $input) {
            clientMutationId
            workItem {
                id
            }
          }
        }
    """
    assigned_users = ["user1", "user2"]
    meta = {"test": "test"}
    inp = {
        "input": {
            "case": str(work_item.case.pk),
            "multipleInstanceTask": str(work_item.task.pk),
            "assignedUsers": assigned_users,
            "meta": json.dumps(meta),
            "description": "work_item description",
        }
    }

    if set_groups:
        inp["input"]["controllingGroups"] = []
        inp["input"]["addressedGroups"] = []

    result = schema_executor(query, variable_values=inp, info=info)

    assert not bool(result.errors) == success
    if success:
        pk = extract_global_id(result.data["createWorkItem"]["workItem"]["id"])
        new_work_item = models.WorkItem.objects.get(pk=pk)
        assert new_work_item.assigned_users == assigned_users
        assert new_work_item.status == models.WorkItem.STATUS_READY
        assert new_work_item.meta == meta
        assert new_work_item.document is not None
        assert new_work_item.controlling_groups == expected_groups
        assert new_work_item.addressed_groups == expected_groups
        assert dict(new_work_item.name) == {"en": "Task name", "de": None, "fr": None}
        assert dict(new_work_item.description) == {
            "en": "work_item description",
            "de": None,
            "fr": None,
        }


def test_filter_document_has_answer(
    db, schema_executor, simple_case, work_item_factory
):
    item_a, item_b = work_item_factory.create_batch(2)

    item_a.document = simple_case.document
    item_a.save()
    item_b.case = simple_case
    item_b.save()

    answer = simple_case.document.answers.first()
    answer.value = "foo"
    answer.save()
    answer.question.type = answer.question.TYPE_TEXT
    answer.question.save()

    query_expect = [("documentHasAnswer", item_a), ("caseDocumentHasAnswer", item_b)]

    for filt, expected in query_expect:
        query = """
            query WorkItems($has_answer: [HasAnswerFilterType!]) {
              allWorkItems(%(filt)s: $has_answer) {
                edges {
                  node {
                    id
                  }
                }
              }
            }
        """ % {
            "filt": filt
        }

        result = schema_executor(
            query,
            variable_values={
                "has_answer": [{"question": answer.question_id, "value": "foo"}]
            },
        )

        assert not result.errors
        assert len(result.data["allWorkItems"]["edges"]) == 1
        node_id = extract_global_id(
            result.data["allWorkItems"]["edges"][0]["node"]["id"]
        )
        assert node_id == str(expected.id)


@pytest.mark.parametrize(
    "lookup_expr,int,value,len_results",
    [
        (None, False, "value1", 1),
        ("EXACT", False, "value1", 1),
        ("EXACT", False, "value", 0),
        ("STARTSWITH", False, "alue1", 0),
        ("STARTSWITH", False, "val", 2),
        ("CONTAINS", False, "alue", 2),
        ("CONTAINS", False, "AlUe", 0),
        ("ICONTAINS", False, "AlUe", 2),
        ("GTE", True, 2, 2),
        ("GTE", True, 5, 1),
        ("GTE", True, 6, 0),
        ("GT", True, 4, 1),
        ("GT", True, 5, 0),
        ("LTE", True, 1, 0),
        ("LTE", True, 2, 1),
        ("LTE", True, 6, 2),
        ("LT", True, 2, 0),
        ("LT", True, 6, 2),
    ],
)
def test_query_all_work_items_filter_case_meta_value(
    db, work_item_factory, schema_executor, lookup_expr, int, value, len_results
):
    work_item_factory(case__meta={"testkey": 2 if int else "value1"})
    work_item_factory(case__meta={"testkey": 5 if int else "value2"})
    work_item_factory(case__meta={"testkey2": 2 if int else "value1"})

    query = """
        query WorkItems($case_meta_value: [JSONValueFilterType!]) {
          allWorkItems(caseMetaValue: $case_meta_value) {
            totalCount
            edges {
              node {
                case {
                  meta
                }
              }
            }
          }
        }
    """

    variables = {"key": "testkey", "value": value}

    if lookup_expr:
        variables["lookup"] = lookup_expr

    result = schema_executor(query, variable_values={"case_meta_value": [variables]})

    assert not result.errors
    assert len(result.data["allWorkItems"]["edges"]) == len_results


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,case__status,task__type,case__document",
    [
        (
            models.WorkItem.STATUS_READY,
            None,
            models.Case.STATUS_RUNNING,
            models.Task.TYPE_COMPLETE_TASK_FORM,
            None,
        )
    ],
)
@pytest.mark.parametrize(
    "question__type,answer__value",
    [(Question.TYPE_INTEGER, 1), (Question.TYPE_CHOICE, "")],
)
def test_skip_task_form_work_item(
    db, work_item, answer, form_question, schema_executor
):
    query = """
        mutation SkipWorkItem($input: SkipWorkItemInput!) {
          skipWorkItem(input: $input) {
            workItem {
              status
              case {
                status
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    # when skipping, both valid and invalid forms don't matter (contrary
    # to completion, where it DOES matter)
    assert not bool(result.errors)
    assert result.data["skipWorkItem"]["workItem"]["status"] == to_const(
        models.WorkItem.STATUS_SKIPPED
    )
    assert result.data["skipWorkItem"]["workItem"]["case"]["status"] == to_const(
        models.Case.STATUS_COMPLETED
    )


@pytest.mark.parametrize(
    "work_item_status,success",
    [
        (models.WorkItem.STATUS_COMPLETED, False),
        (models.WorkItem.STATUS_CANCELED, False),
        (models.WorkItem.STATUS_SKIPPED, False),
        (models.WorkItem.STATUS_READY, True),
    ],
)
@pytest.mark.parametrize("question__type,answer__value", [(Question.TYPE_INTEGER, 1)])
def test_skip_multiple_instance_task_form_work_item(
    db,
    task_factory,
    work_item_factory,
    answer,
    form_question,
    work_item_status,
    success,
    schema_executor,
):
    task = task_factory(is_multiple_instance=True)
    work_item_1 = work_item_factory(task=task, child_case=None, status=work_item_status)
    work_item_2 = work_item_factory(task=task, child_case=None, case=work_item_1.case)
    query = """
        mutation SkipWorkItem($input: SkipWorkItemInput!) {
          skipWorkItem(input: $input) {
            workItem {
              status
              case {
                status
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item_1.pk}}
    result = schema_executor(query, variable_values=inp)

    assert bool(result.errors) != success

    if not success:
        assert "Only READY work items can be skipped" in str(result.errors[0])

    else:

        assert result.data["skipWorkItem"]["workItem"]["status"] == to_const(
            models.WorkItem.STATUS_SKIPPED
        )
        assert result.data["skipWorkItem"]["workItem"]["case"]["status"] == to_const(
            models.Case.STATUS_RUNNING
        )

        inp = {"input": {"id": work_item_2.pk}}
        result = schema_executor(query, variable_values=inp)

        assert not bool(result.errors)
        assert result.data["skipWorkItem"]["workItem"]["status"] == to_const(
            models.WorkItem.STATUS_SKIPPED
        )
        assert result.data["skipWorkItem"]["workItem"]["case"]["status"] == to_const(
            models.Case.STATUS_COMPLETED
        )


@pytest.mark.parametrize("task__type,task__form", [(models.Task.TYPE_SIMPLE, None)])
@pytest.mark.parametrize(
    "work_item__status,case__status",
    [(models.WorkItem.STATUS_READY, models.Case.STATUS_COMPLETED)],
)
def test_complete_work_item_last_api(db, work_item, admin_user):
    api.complete_work_item(work_item, admin_user)

    work_item.refresh_from_db()

    assert work_item.status == models.WorkItem.STATUS_COMPLETED
    assert work_item.case.status == models.Case.STATUS_COMPLETED


@pytest.mark.parametrize("task__type,task__form", [(models.Task.TYPE_SIMPLE, None)])
@pytest.mark.parametrize(
    "work_item__status,case__status",
    [(models.WorkItem.STATUS_READY, models.Case.STATUS_COMPLETED)],
)
def test_skip_work_item_last_api(db, work_item, admin_user):
    api.skip_work_item(work_item, admin_user)

    work_item.refresh_from_db()

    assert work_item.status == models.WorkItem.STATUS_SKIPPED
    assert work_item.case.status == models.Case.STATUS_COMPLETED


@pytest.mark.parametrize(
    "work_item__status,work_item__child_case,task__type",
    [(models.WorkItem.STATUS_COMPLETED, None, models.Task.TYPE_SIMPLE)],
)
def test_complete_work_item_parallel(
    db,
    case,
    work_item,
    work_item_factory,
    task,
    task_factory,
    flow_factory,
    task_flow_factory,
    workflow,
    schema_executor,
    admin_user,
):
    # create two work items which can be processed in parallel
    task_1, task_2 = task_factory.create_batch(2, type=models.Task.TYPE_SIMPLE)
    flow = flow_factory(next=f"['{task_1.slug}','{task_2.slug}']|tasks")
    task_flow_factory(task=work_item.task, workflow=workflow, flow=flow)
    work_item_factory(
        task=task_1, status=models.WorkItem.STATUS_READY, child_case=None, case=case
    )
    work_item_factory(
        task=task_2, status=models.WorkItem.STATUS_READY, child_case=None, case=case
    )

    ready_work_items = case.work_items.filter(status=models.WorkItem.STATUS_READY)

    for i, ready_work_item in enumerate(ready_work_items):
        api.complete_work_item(ready_work_item, admin_user)

        ready_work_item.refresh_from_db()
        case.refresh_from_db()

        assert ready_work_item.status == models.WorkItem.STATUS_COMPLETED

        # if we have multiple parallel running work items without a flow (end
        # of the workflow), the case should only be completed when all parallel
        # work items are completed
        if i + 1 == len(ready_work_items):
            assert case.status == models.Case.STATUS_COMPLETED
        else:
            assert case.status == models.Case.STATUS_RUNNING


def test_complete_work_item_same_task_multiple_workflows(
    db,
    case_factory,
    work_item_factory,
    task_factory,
    flow_factory,
    task_flow_factory,
    workflow_factory,
    schema_executor,
    admin_user,
):
    workflow_1, workflow_2 = workflow_factory.create_batch(2)
    # create two work items which can be processed in parallel
    task_1, task_2 = task_factory.create_batch(2, type=models.Task.TYPE_SIMPLE)

    flow = flow_factory(next=f"'{task_2.slug}'|task")

    # workflow 1 consists out of 2 tasks, workflow_2 just out of one
    task_flow_factory(task=task_1, workflow=workflow_1, flow=flow)

    case_1 = case_factory(workflow=workflow_1)
    case_2 = case_factory(workflow=workflow_2)

    work_item_1 = work_item_factory(
        task=task_1, status=models.WorkItem.STATUS_READY, child_case=None, case=case_1
    )
    work_item_2 = work_item_factory(
        task=task_1, status=models.WorkItem.STATUS_READY, child_case=None, case=case_2
    )

    api.complete_work_item(work_item_1, admin_user)
    api.complete_work_item(work_item_2, admin_user)

    work_item_1.refresh_from_db()
    work_item_2.refresh_from_db()

    case_1.refresh_from_db()
    case_2.refresh_from_db()

    assert case_1.status == models.Case.STATUS_RUNNING
    assert case_2.status == models.Case.STATUS_COMPLETED



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-21 15:17
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="answer",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="answer",
            name="meta",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="document",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="document",
            name="meta",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="form",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="form",
            name="meta",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="formquestion",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="meta",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="question",
            name="configuration",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="question",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="meta",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="questionoption",
            name="created_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
    ]



import pytest

from ...caluma_core.relay import extract_global_id


@pytest.mark.parametrize(
    "lookup,search,expect",
    [
        (None, "bar", ["exact"]),
        ("EXACT", "bar", ["exact"]),
        ("STARTSWITH", "bar", ["exact", "starts"]),
        ("CONTAINS", "bar", ["contains", "starts", "exact"]),
        ("ICONTAINS", "bar", ["icontains", "contains", "starts", "exact"]),
        (None, True, ["bool"]),
        (None, 123, ["int"]),
        (None, 123.456, ["float"]),
    ],
)
def test_meta_value_filter(
    db, schema_executor, document_factory, lookup, search, expect
):
    docs = {
        "nope": document_factory(meta={"foo": "does not match"}),
        "exact": document_factory(meta={"foo": "bar"}),
        "starts": document_factory(meta={"foo": "bar is what it starts with"}),
        "contains": document_factory(meta={"foo": "contains a bar somewhere"}),
        "icontains": document_factory(meta={"foo": "contains a Bar somewhere"}),
        "bool": document_factory(meta={"foo": True}),
        "int": document_factory(meta={"foo": 123}),
        "float": document_factory(meta={"foo": 123.456}),
    }

    query = """
        query AllDocumentsQuery($filter: [JSONValueFilterType]) {
          allDocuments(metaValue: $filter) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    variables = {"filter": [{"key": "foo", "value": search}]}
    if lookup:
        variables["filter"][0]["lookup"] = lookup

    result = schema_executor(query, variable_values=variables)

    assert not result.errors

    expected_ids = [str(docs[exp].id) for exp in expect]

    received_ids = [
        extract_global_id(entry["node"]["id"])
        for entry in result.data["allDocuments"]["edges"]
    ]

    assert sorted(received_ids) == sorted(expected_ids)



from typing import Optional

from caluma.caluma_form.models import Form
from caluma.caluma_user.models import BaseUser

from . import domain_logic, models


def start_case(
    workflow: models.Workflow,
    user: BaseUser,
    form: Optional[Form] = None,
    parent_work_item: Optional[models.WorkItem] = None,
    **kwargs
) -> models.Case:
    """
    Start a case of a given workflow (just like `saveCase`).

    Similar to Case.objects.create(), but with the same business logic as
    used in the `saveCase` mutation.

    >>> start_case(
    ...     workflow=Workflow.objects.get(pk="my-wf"),
    ...     form=Form.objects.get(pk="my-form")),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    data = {"workflow": workflow, "form": form, "parent_work_item": parent_work_item}

    data.update(kwargs)

    validated_data = domain_logic.StartCaseLogic.pre_start(
        domain_logic.StartCaseLogic.validate_for_start(data), user
    )

    case = models.Case.objects.create(**validated_data)

    return domain_logic.StartCaseLogic.post_start(case, user, parent_work_item)


def complete_work_item(work_item: models.WorkItem, user: BaseUser) -> models.WorkItem:
    """
    Complete a work item (just like `completeWorkItem`).

    >>> complete_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    ```
    """
    domain_logic.CompleteWorkItemLogic.validate_for_complete(work_item, user)
    validated_data = domain_logic.CompleteWorkItemLogic.pre_complete({}, user)

    models.WorkItem.objects.filter(pk=work_item.pk).update(**validated_data)

    domain_logic.CompleteWorkItemLogic.post_complete(work_item, user)

    return work_item


def skip_work_item(work_item: models.WorkItem, user: BaseUser) -> models.WorkItem:
    """
    Skip a work item (just like `SkipWorkItem`).

    >>> skip_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.SkipWorkItemLogic.validate_for_skip(work_item)

    validated_data = domain_logic.SkipWorkItemLogic.pre_skip({}, user)

    models.WorkItem.objects.filter(pk=work_item.pk).update(**validated_data)

    domain_logic.SkipWorkItemLogic.post_skip(work_item, user)

    return work_item


def cancel_case(case: models.Case, user: BaseUser) -> models.Case:
    """
    Cancel a case and its pending work items (just like `CancelCase`).

    >>> cancel_case(
    ...     case=models.Case.first(),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    domain_logic.CancelCaseLogic.validate_for_cancel(case)

    validated_data = domain_logic.CancelCaseLogic.pre_cancel({}, user)

    models.Case.objects.filter(pk=case.pk).update(**validated_data)

    domain_logic.CancelCaseLogic.post_cancel(case, user)

    return case



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_fetch_data_sources 1"] = {
    "allDataSources": {
        "edges": [
            {"node": {"info": "Nice test data source", "name": "MyDataSource"}},
            {"node": {"info": "Faulty test data source", "name": "MyFaultyDataSource"}},
            {
                "node": {
                    "info": "Other faulty test data source",
                    "name": "MyOtherFaultyDataSource",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_data_sources 2"] = {
    "allDataSources": {
        "edges": [
            {"node": {"info": "Schöne Datenquelle", "name": "MyDataSource"}},
            {"node": {"info": "Faulty test data source", "name": "MyFaultyDataSource"}},
            {
                "node": {
                    "info": "Other faulty test data source",
                    "name": "MyOtherFaultyDataSource",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_data_sources 3"] = {
    "allDataSources": {
        "edges": [
            {"node": {"info": "Nice test data source", "name": "MyDataSource"}},
            {"node": {"info": "Faulty test data source", "name": "MyFaultyDataSource"}},
            {
                "node": {
                    "info": "Other faulty test data source",
                    "name": "MyOtherFaultyDataSource",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_data_from_data_source 1"] = {
    "dataSource": {
        "edges": [
            {"node": {"label": "1", "slug": "1"}},
            {"node": {"label": "5.5", "slug": "5.5"}},
            {"node": {"label": "sdkj", "slug": "sdkj"}},
            {"node": {"label": "info", "slug": "value"}},
            {"node": {"label": "something", "slug": "something"}},
            {"node": {"label": "english description", "slug": "translated_value"}},
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjU=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 6,
    }
}

snapshots["test_data_source_defaults 1"] = {
    "dataSource": {
        "edges": [
            {"node": {"label": "1", "slug": "1"}},
            {"node": {"label": "2", "slug": "2"}},
            {"node": {"label": "3", "slug": "3"}},
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_data_from_data_source 2"] = {
    "dataSource": {
        "edges": [
            {"node": {"label": "1", "slug": "1"}},
            {"node": {"label": "5.5", "slug": "5.5"}},
            {"node": {"label": "sdkj", "slug": "sdkj"}},
            {"node": {"label": "info", "slug": "value"}},
            {"node": {"label": "something", "slug": "something"}},
            {"node": {"label": "deutsche Beschreibung", "slug": "translated_value"}},
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjU=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 6,
    }
}

snapshots["test_fetch_data_from_data_source 3"] = {
    "dataSource": {
        "edges": [
            {"node": {"label": "1", "slug": "1"}},
            {"node": {"label": "5.5", "slug": "5.5"}},
            {"node": {"label": "sdkj", "slug": "sdkj"}},
            {"node": {"label": "info", "slug": "value"}},
            {"node": {"label": "something", "slug": "something"}},
            {"node": {"label": "english description", "slug": "translated_value"}},
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjU=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 6,
    }
}



# Generated by Django 2.2.10 on 2020-03-05 09:17

import django.db.models.deletion
from django.db import migrations, models


def set_family(apps, schema_editor):
    Case = apps.get_model("caluma_workflow", "Case")
    HistoricalCase = apps.get_model("caluma_workflow", "HistoricalCase")
    db_alias = schema_editor.connection.alias

    orphan_cases = Case.objects.using(db_alias).filter(family__isnull=True)
    orphan_historical_cases = HistoricalCase.objects.using(db_alias).filter(
        family__isnull=True
    )

    for case in orphan_cases:
        family = case
        while hasattr(family, "parent_work_item"):
            family = case.parent_work_item.case
        case.family = family

    Case.objects.using(db_alias).bulk_update(orphan_cases, ["family"])

    for case in orphan_historical_cases:
        # for historical cases, it's possible that the original case doesn't exist
        # anymore. In those cases we set the family to the id of the (now non-existent)
        # original case.
        family = case
        try:
            family = Case.objects.using(db_alias).get(pk=family.id).family
        except Case.DoesNotExist:
            pass
        case.family_id = family.id

    HistoricalCase.objects.using(db_alias).bulk_update(
        orphan_historical_cases, ["family"]
    )


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0019_slugfield_length")]

    operations = [
        migrations.AddField(
            model_name="case",
            name="family",
            field=models.ForeignKey(
                help_text="Family id which case belongs to.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_workflow.Case",
            ),
        ),
        migrations.AddField(
            model_name="historicalcase",
            name="family",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                help_text="Family id which case belongs to.",
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_workflow.Case",
            ),
        ),
        migrations.RunPython(set_family, migrations.RunPython.noop),
    ]



from django.dispatch import receiver
from simple_history.signals import pre_create_historical_record

from . import models


@receiver(pre_create_historical_record, sender=models.HistoricalAnswer)
def set_historical_answer_type(sender, **kwargs):
    """Receiver for historical answer pre_save signal setting question type."""
    historical_answer = kwargs["history_instance"]

    # can take the current question, assuming the question type is not changed in a datarace
    historical_answer.history_question_type = historical_answer.question.type



import pytest
from graphql_relay import to_global_id

from ...caluma_core.relay import extract_global_id
from ...caluma_core.tests import extract_serializer_input_fields
from ...caluma_core.visibilities import BaseVisibility, filter_queryset_for
from ...caluma_form.models import Answer, Document, Question
from ...caluma_form.schema import Document as DocumentNodeType
from .. import serializers


@pytest.mark.parametrize(
    "question__type,question__data_source,answer__value,answer__date",
    [
        (Question.TYPE_INTEGER, None, 1, None),
        (Question.TYPE_FLOAT, None, 2.1, None),
        (Question.TYPE_TEXT, None, "somevalue", None),
        (Question.TYPE_MULTIPLE_CHOICE, None, ["somevalue", "anothervalue"], None),
        (Question.TYPE_TABLE, None, None, None),
        (Question.TYPE_DATE, None, None, "2019-02-22"),
        (Question.TYPE_FILE, None, "some-file.pdf", None),
        (Question.TYPE_FILE, None, "some-other-file.pdf", None),
        (Question.TYPE_DYNAMIC_CHOICE, "MyDataSource", "5.5", None),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, "MyDataSource", ["5.5"], None),
    ],
)
def test_query_all_documents(
    db,
    snapshot,
    form_question,
    form,
    document,
    document_factory,
    question_factory,
    form_question_factory,
    answer_factory,
    file_factory,
    answer_document,
    answer,
    schema_executor,
    question,
    minio_mock,
    data_source_settings,
    settings,
):
    query = """
        query AllDocumentsQuery($search: String) {
          allDocuments(search: $search) {
            totalCount
            edges {
              node {
                createdByUser
                answers {
                  totalCount
                  edges {
                    node {
                      __typename
                      question {
                        slug
                        label
                      }
                      ... on StringAnswer {
                        string_value: value
                      }
                      ... on IntegerAnswer {
                        integer_value: value
                      }
                      ... on ListAnswer {
                        list_value: value
                      }
                      ... on FloatAnswer {
                        float_value: value
                      }
                      ... on DateAnswer {
                        date_value: value
                      }
                      ... on TableAnswer {
                        table_value: value {
                          form {
                            slug
                          }
                        }
                      }
                      ... on FileAnswer {
                        fileValue: value {
                          name
                          downloadUrl
                          metadata
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
    """

    search = isinstance(answer.value, list) and " ".join(answer.value) or answer.value

    if question.type == Question.TYPE_FILE:
        if answer.value == "some-other-file.pdf":
            settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = False
            minio_mock.bucket_exists.return_value = False
        # we need to set the pk here in order to match the snapshots
        answer.file = file_factory(
            name=answer.value, pk="09c697fb-fd0a-4345-bb9c-99df350b0cdb"
        )
        answer.value = None
        answer.save()
        search = answer.file.name

    result = schema_executor(query, variable_values={"search": search})
    assert not result.errors
    snapshot.assert_match(result.data)


def test_complex_document_query_performance(
    db,
    schema_executor,
    document,
    form,
    form_question_factory,
    question_factory,
    answer_factory,
    file_factory,
    question_option_factory,
    django_assert_num_queries,
    minio_mock,
):

    answers = answer_factory.create_batch(5, document=document)
    for answer in answers:
        form_question_factory(question=answer.question, form=form)
    multiple_choice_question = question_factory(type=Question.TYPE_MULTIPLE_CHOICE)
    form_question_factory(question=multiple_choice_question, form=form)
    question_option_factory.create_batch(10, question=multiple_choice_question)
    answer_factory(question=multiple_choice_question)
    file_question = question_factory(type=Question.TYPE_FILE)
    form_question_factory(question=file_question, form=form)
    answer_factory(
        question=file_question, value=None, document=document, file=file_factory()
    )

    form_question = question_factory(type=Question.TYPE_FORM)
    form_question_factory(question=form_question, form=form)

    table_question = question_factory(type=Question.TYPE_TABLE)
    form_question_factory(question=table_question, form=form)

    query = """
        query ($id: ID!) {
          allDocuments(id: $id) {
            edges {
              node {
                ...FormDocument
              }
            }
          }
        }

        fragment FormDocument on Document {
          id
          answers {
            edges {
              node {
                ...FieldAnswer
              }
            }
          }
          form {
            slug
            questions {
              edges {
                node {
                  ...FieldQuestion
                }
              }
            }
          }
        }

        fragment FieldAnswer on Answer {
          id
          __typename
          question {
            slug
          }
          ... on StringAnswer {
            stringValue: value
          }
          ... on IntegerAnswer {
            integerValue: value
          }
          ... on FloatAnswer {
            floatValue: value
          }
          ... on DateAnswer {
            dateValue: value
          }
          ... on ListAnswer {
            listValue: value
          }
          ... on FileAnswer {
            fileValue: value {
              name
              downloadUrl
              metadata
            }
          }
        }

        fragment FieldQuestion on Question {
          slug
          label
          isRequired
          isHidden
          ... on TextQuestion {
            textMaxLength: maxLength
          }
          ... on TextareaQuestion {
            textareaMaxLength: maxLength
          }
          ... on IntegerQuestion {
            integerMinValue: minValue
            integerMaxValue: maxValue
          }
          ... on FloatQuestion {
            floatMinValue: minValue
            floatMaxValue: maxValue
          }
          ... on ChoiceQuestion {
            choiceOptions: options {
              edges {
                node {
                  slug
                  label
                }
              }
            }
          }
          ... on FormQuestion {
            subForm {
              slug
              name
            }
          }
          ... on TableQuestion {
            rowForm {
              slug
            }
          }
          ... on MultipleChoiceQuestion {
            multipleChoiceOptions: options {
              edges {
                node {
                  slug
                  label
                }
              }
            }
          }
        }
    """

    with django_assert_num_queries(8):
        result = schema_executor(query, variable_values={"id": str(document.pk)})
    assert not result.errors


def test_query_all_documents_filter_answers_by_question(
    db, document, answer, question, answer_factory, schema_executor
):
    answer_factory(document=document)

    query = """
        query AllDocumentsQuery($question: ID!) {
          allDocuments {
            edges {
              node {
                answers(question: $question) {
                  edges {
                    node {
                      id
                    }
                  }
                }
              }
            }
          }
        }
    """

    result = schema_executor(query, variable_values={"question": question.slug})
    assert not result.errors
    assert len(result.data["allDocuments"]["edges"]) == 1
    result_document = result.data["allDocuments"]["edges"][0]["node"]
    assert len(result_document["answers"]["edges"]) == 1
    result_answer = result_document["answers"]["edges"][0]["node"]
    assert extract_global_id(result_answer["id"]) == str(answer.id)


def test_query_all_documents_filter_answers_by_questions(
    db, document_factory, question_factory, answer_factory, schema_executor
):
    documents = []
    answers = []
    questions = []

    for _ in range(3):
        documents.append(document_factory())
        questions.append(question_factory())
        answers.append(answer_factory(document=documents[-1], question=questions[-1]))

    query = """
        query AllDocumentsQuery($questions: [ID!]) {
          allDocuments {
            edges {
              node {
                id
                answers(questions: $questions) {
                  edges {
                    node {
                      id
                    }
                  }
                }
              }
            }
          }
        }
    """

    result = schema_executor(
        query, variable_values={"questions": [questions[0].slug, questions[1].slug]}
    )
    assert not result.errors
    assert len(result.data["allDocuments"]["edges"]) == 3

    result_lengths = [
        (extract_global_id(doc["node"]["id"]), len(doc["node"]["answers"]["edges"]))
        for doc in result.data["allDocuments"]["edges"]
    ]
    expect_data = [(str(documents[idx].pk), int(idx < 2)) for idx in range(3)]
    assert set(expect_data) == set(result_lengths)


@pytest.mark.parametrize("update", [True, False])
def test_save_document(db, document, schema_executor, update):
    query = """
        mutation SaveDocument($input: SaveDocumentInput!) {
          saveDocument(input: $input) {
            document {
                form {
                  slug
                }
                id
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.DocumentSerializer, document
        )
    }

    if not update:
        # not update = create = we don't pass the ID
        del inp["input"]["id"]

    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    assert result.data["saveDocument"]["document"]["form"]["slug"] == document.form.slug

    same_id = extract_global_id(result.data["saveDocument"]["document"]["id"]) == str(
        document.id
    )

    # if updating, the resulting document must be the same
    assert same_id == update


@pytest.mark.parametrize("delete_answer", [True, False])
@pytest.mark.parametrize("option__slug", ["option-slug"])
@pytest.mark.parametrize(
    "question__type,question__configuration,question__data_source,question__format_validators,answer__value,answer__date,mutation,success",
    [
        (
            Question.TYPE_INTEGER,
            {},
            None,
            [],
            1,
            None,
            "SaveDocumentIntegerAnswer",
            True,
        ),
        (
            Question.TYPE_INTEGER,
            {"min_value": 100},
            None,
            [],
            1,
            None,
            "SaveDocumentIntegerAnswer",
            False,
        ),
        (Question.TYPE_FLOAT, {}, None, [], 2.1, None, "SaveDocumentFloatAnswer", True),
        (
            Question.TYPE_FLOAT,
            {"min_value": 100.0},
            None,
            [],
            1,
            None,
            "SaveDocumentFloatAnswer",
            False,
        ),
        (
            Question.TYPE_TEXT,
            {},
            None,
            [],
            "Test",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
        (
            Question.TYPE_TEXT,
            {"max_length": 1},
            None,
            [],
            "toolong",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_DATE,
            {},
            None,
            [],
            None,
            "1900-01-01",
            "SaveDocumentDateAnswer",
            False,
        ),
        (
            Question.TYPE_DATE,
            {},
            None,
            [],
            None,
            "2019-02-22",
            "SaveDocumentDateAnswer",
            True,
        ),
        (Question.TYPE_FILE, {}, None, [], None, None, "SaveDocumentFileAnswer", False),
        (
            Question.TYPE_FILE,
            {},
            None,
            [],
            "some-file.pdf",
            None,
            "SaveDocumentFileAnswer",
            True,
        ),
        (
            Question.TYPE_FILE,
            {},
            None,
            [],
            "not-exist.pdf",
            None,
            "SaveDocumentFileAnswer",
            True,
        ),
        (
            Question.TYPE_TEXT,
            {"min_length": 10},
            None,
            [],
            "tooshort",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_TABLE,
            {},
            None,
            [],
            None,
            None,
            "SaveDocumentTableAnswer",
            True,
        ),
        (
            Question.TYPE_TEXTAREA,
            {},
            None,
            [],
            "Test",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
        (
            Question.TYPE_TEXTAREA,
            {"max_length": 1},
            None,
            [],
            "toolong",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_MULTIPLE_CHOICE,
            {},
            None,
            [],
            ["option-slug"],
            None,
            "SaveDocumentListAnswer",
            True,
        ),
        (
            Question.TYPE_MULTIPLE_CHOICE,
            {},
            None,
            [],
            ["option-slug", "option-invalid-slug"],
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_CHOICE,
            {},
            None,
            [],
            "option-slug",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
        (
            Question.TYPE_CHOICE,
            {},
            None,
            [],
            "invalid-option-slug",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            {},
            "MyDataSource",
            [],
            ["5.5", "1"],
            None,
            "SaveDocumentListAnswer",
            True,
        ),
        (
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            {},
            "MyDataSource",
            [],
            ["not in data"],
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_DYNAMIC_CHOICE,
            {},
            "MyDataSource",
            [],
            "5.5",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
        (
            Question.TYPE_DYNAMIC_CHOICE,
            {},
            "MyDataSource",
            [],
            "not in data",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_TEXT,
            {},
            None,
            ["email"],
            "some text",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_TEXT,
            {},
            None,
            ["email"],
            "test@example.com",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
        (
            Question.TYPE_TEXTAREA,
            {},
            None,
            ["email"],
            "some text",
            None,
            "SaveDocumentStringAnswer",
            False,
        ),
        (
            Question.TYPE_TEXTAREA,
            {},
            None,
            ["email"],
            "test@example.com",
            None,
            "SaveDocumentStringAnswer",
            True,
        ),
    ],
)
def test_save_document_answer(
    db,
    snapshot,
    question,
    answer,
    mutation,
    question_option,
    document_factory,
    answer_factory,
    answer_document_factory,
    question_factory,
    file_factory,
    success,
    schema_executor,
    delete_answer,
    minio_mock,
    data_source_settings,
):
    mutation_func = mutation[0].lower() + mutation[1:]
    query = f"""
        mutation {mutation}($input: {mutation}Input!) {{
          {mutation_func}(input: $input) {{
            answer {{
              __typename
              ... on StringAnswer {{
                stringValue: value
              }}
              ... on IntegerAnswer {{
                integerValue: value
              }}
              ... on ListAnswer {{
                listValue: value
              }}
              ... on FloatAnswer {{
                floatValue: value
              }}
              ... on ListAnswer {{
                listValue: value
              }}
              ... on DateAnswer {{
                dateValue: value
              }}
              ... on TableAnswer {{
                table_value: value {{
                  form {{
                    slug
                  }}
                }}
              }}
              ... on FileAnswer {{
                fileValue: value {{
                  name
                  uploadUrl
                }}
              }}
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveAnswerSerializer, answer
        )
    }

    if question.type == Question.TYPE_TABLE:
        documents = document_factory.create_batch(2, form=question.row_form)
        # create a subtree
        sub_question = question_factory(type=Question.TYPE_TEXT)
        document_answer = answer_factory(question=sub_question)
        documents[0].answers.add(document_answer)
        answer_document_factory(answer=answer, document=documents[0])

        inp["input"]["value"] = [str(document.pk) for document in documents]

    if question.type == Question.TYPE_FILE:
        if answer.value == "some-file.pdf":
            minio_mock.bucket_exists.return_value = False
        answer.value = None
        answer.save()

    if question.type == Question.TYPE_DATE:
        inp["input"]["value"] = answer.date
        answer.value = None
        answer.save()
        # Date format is enforced in the model. So we initially had to use a valid date
        # here we're able to change it:
        if answer.date == "1900-01-01":
            inp["input"]["value"] = "not a date"

    if delete_answer:
        # delete answer to force create test instead of update
        Answer.objects.filter(pk=answer.pk).delete()

    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete_answer", [True, False])
@pytest.mark.parametrize(
    "question__type,question__is_required,answer__value",
    [(Question.TYPE_TEXT, "false", None)],
)
def test_save_document_answer_empty(
    db, snapshot, question, answer, schema_executor, delete_answer
):
    query = f"""
        mutation saveDocumentStringAnswer($input: SaveDocumentStringAnswerInput!) {{
          saveDocumentStringAnswer(input: $input) {{
            answer {{
              __typename
              ... on StringAnswer {{
                stringValue: value
              }}
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": {
            "document": to_global_id("StringAnswer", answer.document.pk),
            "question": to_global_id("StringAnswer", answer.question.pk),
        }
    }

    if delete_answer:
        # delete answer to force create test instead of update
        Answer.objects.filter(pk=answer.pk).delete()

    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [Question.TYPE_TABLE])
def test_save_document_table_answer_invalid_row_document(
    db,
    schema_executor,
    answer,
    document_factory,
    question,
    form_factory,
    question_factory,
):
    """Ensure that we can save incomplete row documents."""
    question.row_form = form_factory()
    question.save()

    question.row_form.questions.add(question_factory(is_required="true"))

    query = """
        mutation SaveDocumentTableAnswer($input: SaveDocumentTableAnswerInput!) {
            saveDocumentTableAnswer(input: $input) {
                clientMutationId
            }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveAnswerSerializer, answer
        )
    }
    inp["input"]["value"] = [str(document_factory(form=question.row_form).pk)]

    result = schema_executor(query, variable_values=inp)
    assert not result.errors


@pytest.mark.parametrize("question__type", [Question.TYPE_TABLE])
def test_save_document_table_answer_invalid_row_form(
    db, schema_executor, answer, document_factory
):
    """Test validation that row documents must have correct row type."""
    query = """
        mutation SaveDocumentTableAnswer($input: SaveDocumentTableAnswerInput!) {
            saveDocumentTableAnswer(input: $input) {
                clientMutationId
            }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveAnswerSerializer, answer
        )
    }
    inp["input"]["value"] = [
        str(document.pk) for document in document_factory.create_batch(1)
    ]
    result = schema_executor(query, variable_values=inp)
    assert result.errors


@pytest.mark.parametrize("question__type", [Question.TYPE_TABLE])
def test_save_document_table_answer_setting_family(
    db,
    schema_executor,
    answer,
    answer_factory,
    document_factory,
    answer_document_factory,
):
    query = """
        mutation SaveDocumentTableAnswer($input: SaveDocumentTableAnswerInput!) {
            saveDocumentTableAnswer(input: $input) {
                clientMutationId
            }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveAnswerSerializer, answer
        )
    }

    main_pk = answer.document.pk
    main_family = answer.document.family
    remaining_document = document_factory(form=answer.question.row_form)
    to_be_deleted_document = document_factory(form=answer.question.row_form)
    answer_factory(document=to_be_deleted_document)
    to_be_deleted_table_row = document_factory(
        form=answer.question.row_form, family=to_be_deleted_document.family
    )
    table_answer = answer_factory(
        question=answer.question, document=to_be_deleted_document
    )
    answer_document_factory(answer=table_answer, document=to_be_deleted_table_row)

    # attach documents to table answer
    inp["input"]["value"] = {remaining_document.pk, to_be_deleted_document.pk}
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    assert {main_pk, to_be_deleted_table_row.pk} | inp["input"]["value"] == set(
        Document.objects.filter(family=main_family).values_list("id", flat=True)
    ), "family is not set to main document"
    to_be_deleted_document.refresh_from_db()
    assert to_be_deleted_document.family == main_family
    to_be_deleted_table_row.refresh_from_db()
    assert to_be_deleted_table_row.family == main_family

    # detach one document answer from table answer
    inp["input"]["value"] = {remaining_document.pk}
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    assert {main_pk} | inp["input"]["value"] == set(
        Document.objects.filter(family=main_family).values_list("id", flat=True)
    )
    to_be_deleted_document.refresh_from_db()
    assert to_be_deleted_document.family == to_be_deleted_document
    to_be_deleted_table_row.refresh_from_db()
    assert to_be_deleted_table_row.family == to_be_deleted_document.family


@pytest.mark.parametrize("answer__value", [1.1])
def test_query_answer_node(db, answer, schema_executor):
    global_id = to_global_id("FloatAnswer", answer.pk)

    node_query = """
    query AnswerNode($id: ID!) {
      node(id: $id) {
        ... on FloatAnswer {
            value
        }
      }
    }
    """

    result = schema_executor(node_query, variable_values={"id": global_id})
    assert not result.errors


@pytest.mark.parametrize(
    "question__is_required,question__type,question__data_source,is_valid",
    [
        ("true", Question.TYPE_TEXT, None, False),
        ("false", Question.TYPE_TEXT, None, True),
        ("true", Question.TYPE_TEXT, None, True),
    ],
)
def test_validity_query(db, form, question, document, is_valid, schema_executor):

    form.questions.through.objects.create(form=form, question=question, sort=1)
    document.form = form
    document.save()

    if is_valid:
        document.answers.create(question=question, value="hello")
    else:
        assert not document.answers.exists()

    query = """
        query ValidateBaugesuch ($document_id: ID!) {
          documentValidity(
            id: $document_id
          ) {
            edges {
              node {
                id
                isValid
                errors {
                  slug
                  errorMsg
                }
              }
            }
          }
        }
    """

    result = schema_executor(query, variable_values={"document_id": document.id})

    # if is_valid, we expect 0 errors, otherwise one
    num_errors = int(not is_valid)

    assert result.data["documentValidity"]["edges"][0]["node"]["id"] == str(document.id)
    assert result.data["documentValidity"]["edges"][0]["node"]["isValid"] == is_valid
    assert (
        len(result.data["documentValidity"]["edges"][0]["node"]["errors"]) == num_errors
    )


@pytest.mark.parametrize("hide_documents", [True, False])
def test_validity_with_visibility(
    db, form, document, schema_executor, hide_documents, mocker
):

    query = """
        query ValidateBaugesuch ($document_id: ID!) {
          documentValidity(
            id: $document_id
          ) {
            edges {
              node {
                id
                isValid
                errors {
                  slug
                  errorMsg
                }
              }
            }
          }
        }
    """

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(DocumentNodeType)
        def filter_queryset_for_document(self, node, queryset, info):
            if hide_documents:
                return queryset.none()
            return queryset

    mocker.patch("caluma.caluma_core.types.Node.visibility_classes", [CustomVisibility])

    result = schema_executor(query, variable_values={"document_id": document.id})

    if hide_documents:
        assert result.data["documentValidity"] is None
    else:
        assert len(result.data["documentValidity"]["edges"]) == 1


def test_remove_document_without_case(db, document, answer, schema_executor):
    query = """
        mutation RemoveDocument($input: RemoveDocumentInput!) {
          removeDocument(input: $input) {
            document {
              id
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query, variable_values={"input": {"document": str(document.pk)}}
    )

    assert not result.errors
    with pytest.raises(Document.DoesNotExist):
        Document.objects.get(pk=document.pk)
    with pytest.raises(Answer.DoesNotExist):
        Answer.objects.get(pk=answer.pk)


def test_remove_document_with_case(db, document, answer, case, schema_executor):
    query = """
        mutation RemoveDocument($input: RemoveDocumentInput!) {
          removeDocument(input: $input) {
            document {
              id
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query, variable_values={"input": {"document": str(document.pk)}}
    )

    assert result.errors
    Document.objects.get(pk=document.pk)
    Answer.objects.get(pk=answer.pk)


def test_remove_document_without_case_table(
    db,
    document_factory,
    answer_factory,
    answer_document_factory,
    question_factory,
    form_question_factory,
    schema_executor,
):
    question = question_factory(type=Question.TYPE_TABLE)
    documents = document_factory.create_batch(2, form=question.row_form)

    sub_question = question_factory(type=Question.TYPE_TEXT)
    form_question_factory(form=question.row_form, question=sub_question)

    sub_answers = answer_factory.create_batch(2, question=sub_question)
    documents[0].answers.add(sub_answers[0])
    documents[1].answers.add(sub_answers[1])

    table_answer = answer_factory(question=question)
    answer_document_factory(answer=table_answer, document=documents[0])
    answer_document_factory(answer=table_answer, document=documents[1])

    query = """
        mutation RemoveDocument($input: RemoveDocumentInput!) {
          removeDocument(input: $input) {
            document {
              id
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query, variable_values={"input": {"document": str(table_answer.document.pk)}}
    )

    assert not result.errors
    for document in [table_answer.document, *documents]:
        with pytest.raises(Document.DoesNotExist):
            Document.objects.get(pk=document.pk)
    for answer in [table_answer, *sub_answers]:
        with pytest.raises(Answer.DoesNotExist):
            Answer.objects.get(pk=answer.pk)


def test_copy_document(
    db,
    document_factory,
    answer_factory,
    answer_document_factory,
    question_factory,
    form_factory,
    form_question_factory,
    schema_executor,
    minio_mock,
):
    main_form = form_factory(slug="main-form")
    table_question = question_factory(
        type=Question.TYPE_TABLE, slug="table-question", row_form=form_factory()
    )
    form_question_factory(form=main_form, question=table_question)

    sub_question = form_question_factory(
        form=table_question.row_form,
        question__type=Question.TYPE_TEXT,
        question__slug="sub_question",
    )
    other_question = form_question_factory(
        form=main_form,
        question__type=Question.TYPE_TEXT,
        question__slug="other_question",
    )
    file_question = form_question_factory(
        form=main_form, question__type=Question.TYPE_FILE
    )

    # main_document
    #   - table_answer
    #       - row_document_1
    #           - sub_question answer:"foo"
    #   - other_question answer:"something"
    #   - file_question answer: b"a file"

    main_document = document_factory(form=main_form)
    table_answer = answer_factory(
        document=main_document, question=table_question, value=None
    )

    row_document_1 = document_factory(form=table_question.row_form)
    answer_document_factory(document=row_document_1, answer=table_answer)

    answer_factory(question=sub_question.question, document=row_document_1, value="foo")
    other_question_ans = answer_factory(
        question=other_question.question, document=main_document, value="something"
    )
    file_answer = answer_factory(
        question=file_question.question, document=main_document
    )

    query = """
        mutation CopyDocument($input: CopyDocumentInput!) {
          copyDocument(input: $input) {
            document {
              id
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query, variable_values={"input": {"source": str(main_document.pk)}}
    )
    assert not result.errors

    result_document_id = extract_global_id(
        result.data["copyDocument"]["document"]["id"]
    )

    new_document = Document.objects.get(pk=result_document_id)
    # main document is copied
    assert new_document.source_id == main_document.pk
    assert new_document.family == new_document
    assert new_document.family != main_document.family

    # answers are copied
    assert other_question.question.slug in new_document.answers.all().values_list(
        "question__slug", flat=True
    )
    assert (
        new_document.answers.get(question__slug=other_question.question.slug).value
        == other_question_ans.value
    )
    assert file_question.question.slug in new_document.answers.all().values_list(
        "question__slug", flat=True
    )
    new_file_answer = new_document.answers.get(
        question__slug=file_question.question.slug
    )
    assert new_file_answer.value == file_answer.value
    # file is copied
    minio_mock.copy_object.assert_called()
    assert new_file_answer.file.name == file_answer.file.name
    assert new_file_answer.file.object_name != file_answer.file.object_name

    # table docs and answers are moved
    new_table_answer = new_document.answers.get(question=table_question)
    assert new_table_answer.documents.count() == 1

    result_table_answer_document = new_table_answer.documents.first()
    assert result_table_answer_document.source == row_document_1
    assert result_table_answer_document.family == new_document
    assert set(ans.value for ans in result_table_answer_document.answers.all()) == set(
        ans.value for ans in row_document_1.answers.all()
    )



from ...caluma_core.relay import extract_global_id
from .. import models


def test_search(
    schema_executor,
    db,
    question,
    form_factory,
    form_question_factory,
    question_factory,
    document_factory,
    answer_factory,
    form,
):

    doc_a, doc_b = document_factory.create_batch(2, form=form)

    question_a = question_factory(type=models.Question.TYPE_TEXT)
    question_b = question_factory(type=models.Question.TYPE_TEXT)

    doc_a.answers.create(question=question_a, value="hello world")
    doc_a.answers.create(question=question_b, value="whatsup planet")

    doc_b.answers.create(question=question_a, value="goodbye planet")
    doc_b.answers.create(question=question_b, value="seeya world")

    query = """
        query ($search: [SearchAnswersFilterType!]) {
          allDocuments (searchAnswers: $search) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    def _search(slugs, word, expect_count):
        variables = {"search": [{"questions": slugs, "value": word}]}
        result = schema_executor(query, variable_values=variables)

        assert not result.errors
        edges = result.data["allDocuments"]["edges"]
        assert len(edges) == expect_count
        return edges

    # search for "hello world". this should return doc a
    edges = _search([question_a.slug], "hello world", 1)
    assert extract_global_id(edges[0]["node"]["id"]) == str(doc_a.id)

    # search for "planet" across both questions. this should return both doc a
    # and b
    edges = _search([question_b.slug, question_a.slug], "planet", 2)
    assert set([str(doc_a.id), str(doc_b.id)]) == set(
        extract_global_id(e["node"]["id"]) for e in edges
    )


def test_search_multiple(
    schema_executor,
    db,
    question,
    form_factory,
    form_question_factory,
    question_factory,
    document_factory,
    answer_factory,
    form,
):

    doc_a, doc_b, doc_c = document_factory.create_batch(3, form=form)

    question_a = question_factory(type=models.Question.TYPE_TEXT)
    question_b = question_factory(type=models.Question.TYPE_TEXT)

    doc_a.answers.create(question=question_a, value="hello planet")
    doc_a.answers.create(question=question_b, value="whatsup planet")

    doc_b.answers.create(question=question_a, value="hello world")
    doc_b.answers.create(question=question_b, value="seeya world")

    doc_c.answers.create(question=question_a, value="goodbye planet")
    doc_c.answers.create(question=question_b, value="seeya world")

    query = """
        query ($search: [SearchAnswersFilterType!]) {
          allDocuments (searchAnswers: $search) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    # Here, we send two searches, which we expect to be joined by AND.
    result = schema_executor(
        query,
        variable_values={
            "search": [
                # "hello" is in doc_a and doc_b
                {"questions": [question_a.slug], "value": "hello"},
                # "world" is in doc_b and doc_c
                {"questions": [question_b.slug], "value": "world"},
            ]
        },
    )

    assert not result.errors
    edges = result.data["allDocuments"]["edges"]

    # Result - we expect only doc_b in the result
    assert len(edges) == 1
    assert extract_global_id(edges[0]["node"]["id"]) == str(doc_b.id)


def test_search_invalid_question_type(schema_executor, db, question_factory):

    question = question_factory(type=models.Question.TYPE_FORM)

    result = schema_executor(
        """
            query ($search: [SearchAnswersFilterType!]) {
              allDocuments (searchAnswers: $search) {
                edges {
                  node {
                    id
                  }
                }
              }
            }
        """,
        variable_values={"search": [{"questions": [question.slug], "value": "blah"}]},
    )

    assert [str(err) for err in result.errors] == [
        ("['Questions of type form cannot be used in searchAnswers']")
    ]



import pytest

from ..relay import extract_global_id


@pytest.mark.parametrize(
    "id,expected",
    [
        ("slug", "slug"),
        ("V29ya2Zsb3dTcGVjaWZpY2F0aW9uOmxpc3RlbmVpbnRyYWc=", "listeneintrag"),
    ],
)
def test_extract_global_id(id, expected):
    assert extract_global_id(id) == expected



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_query_all_forms[First result-1st-float] 1"] = {
    "allForms": {
        "edges": [
            {
                "node": {
                    "description": "First result",
                    "id": "Rm9ybTplZmZvcnQtbWVldA==",
                    "meta": {},
                    "name": "1st",
                    "questions": {
                        "edges": [
                            {
                                "node": {
                                    "id": "RmxvYXRRdWVzdGlvbjphY3Jvc3MtZW52aXJvbm1lbnQ=",
                                    "label": "Tyler Valencia",
                                    "slug": "across-environment",
                                }
                            }
                        ]
                    },
                    "slug": "effort-meet",
                }
            },
            {
                "node": {
                    "description": "Second result",
                    "id": "Rm9ybTpraXRjaGVuLWRldmVsb3A=",
                    "meta": {},
                    "name": "2nd",
                    "questions": {"edges": []},
                    "slug": "kitchen-develop",
                }
            },
            {
                "node": {
                    "description": "Second result",
                    "id": "Rm9ybTpzZXJ2aWNlLWJhbmstYXJt",
                    "meta": {},
                    "name": "3rd",
                    "questions": {"edges": []},
                    "slug": "service-bank-arm",
                }
            },
        ]
    }
}

snapshots["test_save_form[some description text-en] 1"] = {
    "saveForm": {
        "clientMutationId": "testid",
        "form": {
            "id": "Rm9ybTplZmZvcnQtbWVldA==",
            "meta": {},
            "name": "Brian Williams",
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_form[some description text-de] 1"] = {
    "saveForm": {
        "clientMutationId": "testid",
        "form": {
            "id": "Rm9ybTplZmZvcnQtbWVldA==",
            "meta": {},
            "name": "Brian Williams",
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_form[en] 1"] = {
    "saveForm": {
        "clientMutationId": "testid",
        "form": {
            "id": "Rm9ybTplZmZvcnQtbWVldA==",
            "meta": {},
            "name": "Brian Williams",
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_form[de] 1"] = {
    "saveForm": {
        "clientMutationId": "testid",
        "form": {
            "id": "Rm9ybTplZmZvcnQtbWVldA==",
            "meta": {},
            "name": "Brian Williams",
            "slug": "effort-meet",
        },
    }
}

snapshots["test_remove_form_question 1"] = {
    "removeFormQuestion": {
        "clientMutationId": None,
        "form": {"questions": {"edges": []}},
    }
}



import pytest
from django.core.cache import cache
from django.utils import translation

from caluma.caluma_form.models import DynamicOption, Question
from caluma.caluma_user.models import BaseUser


def test_fetch_data_sources(snapshot, schema_executor, settings):
    settings.DATA_SOURCE_CLASSES = [
        "caluma.caluma_data_source.tests.data_sources.MyDataSource",
        "caluma.caluma_data_source.tests.data_sources.MyFaultyDataSource",
        "caluma.caluma_data_source.tests.data_sources.MyOtherFaultyDataSource",
    ]
    query = """
        query dataSources {
          allDataSources {
            totalCount
            pageInfo {
              startCursor
              endCursor
            }
            edges {
              node {
                name
                info
              }
            }
          }
        }
    """

    result = schema_executor(query)
    assert not result.errors
    snapshot.assert_match(result.data)

    with translation.override("de"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)

    with translation.override("nolang"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)


def test_fetch_data_from_data_source(snapshot, schema_executor, data_source_settings):
    query = """
        query dataSource {
          dataSource (name: "MyDataSource") {
            totalCount
            pageInfo {
              startCursor
              endCursor
            }
            edges {
              node {
                label
                slug
              }
            }
          }
        }
    """

    result = schema_executor(query)
    assert not result.errors
    snapshot.assert_match(result.data)
    assert cache.get("data_source_MyDataSource") == [
        1,
        (5.5,),
        "sdkj",
        ("value", "info"),
        ["something"],
        [
            "translated_value",
            {"de": "deutsche Beschreibung", "en": "english description"},
        ],
    ]

    with translation.override("de"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)

    with translation.override("nolang"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "data_source", ["MyFaultyDataSource", "MyOtherFaultyDataSource"]
)
def test_data_source_failure(data_source, schema_executor, settings):
    settings.DATA_SOURCE_CLASSES = [
        f"caluma.caluma_data_source.tests.data_sources.{data_source}"
    ]

    query = """
        query dataSource($name: String!) {
          dataSource (name: $name) {
            pageInfo {
              startCursor
              endCursor
            }
            edges {
              node {
                label
                slug
              }
            }
          }
        }
    """

    inp = {"name": data_source}

    result = schema_executor(query, variable_values=inp)
    assert result.errors


def test_data_source_defaults(snapshot, schema_executor, settings):
    settings.DATA_SOURCE_CLASSES = [
        f"caluma.caluma_data_source.tests.data_sources.MyBrokenDataSource"
    ]

    query = """
            query dataSource {
              dataSource (name: "MyBrokenDataSource") {
                totalCount
                pageInfo {
                  startCursor
                  endCursor
                }
                edges {
                  node {
                    label
                    slug
                  }
                }
              }
            }
        """

    result = schema_executor(query, variable_values={})
    assert not result.errors
    snapshot.assert_match(result.data)


def test_data_source_exception(schema_executor, settings):
    settings.DATA_SOURCE_CLASSES = [
        f"caluma.caluma_data_source.tests.data_sources.MyOtherBrokenDataSource"
    ]

    query = """
            query dataSource {
              dataSource (name: "MyOtherBrokenDataSource") {
                pageInfo {
                  startCursor
                  endCursor
                }
                edges {
                  node {
                    label
                    slug
                  }
                }
              }
            }
        """

    result = schema_executor(query, variable_values={})
    assert result.errors


@pytest.mark.parametrize("configure", [True, False])
def test_fetch_data_from_non_existing_data_source(schema_executor, settings, configure):
    if configure:
        settings.DATA_SOURCE_CLASSES = [
            f"caluma.caluma_data_source.tests.data_sources.NonExistentDataSource"
        ]

    query = """
        query dataSource {
          dataSource (name: "NonExistentDataSource") {
            pageInfo {
              startCursor
              endCursor
            }
            edges {
              node {
                label
                slug
              }
            }
          }
        }
    """

    result = schema_executor(query)
    assert result.errors


@pytest.mark.parametrize(
    "question__type,question__data_source",
    [(Question.TYPE_DYNAMIC_CHOICE, "MyDataSource")],
)
def test_data_sources_stores_user(
    db, schema_executor, info, settings, form, question, document
):
    class FakeUser(BaseUser):
        def __init__(self):
            self.groups = ["foobar"]
            self.username = "asdf"

        @property
        def group(self):
            return "foobar"

        def __str__(self):
            return "asdf"

    settings.DATA_SOURCE_CLASSES = [
        "caluma.caluma_data_source.tests.data_sources.MyDataSource"
    ]
    query = """
      mutation createAnswer($input:SaveDocumentStringAnswerInput!){
        saveDocumentStringAnswer(input:$input){
          answer{
            id
          }
        }
      }
    """
    info.context.user = FakeUser()
    variables = {
        "input": {
            "question": question.slug,
            "document": str(document.id),
            "value": "something",
        }
    }
    assert not DynamicOption.objects.exists()
    result = schema_executor(query, variable_values=variables, info=info)
    assert not result.errors
    assert DynamicOption.objects.filter(
        document=document,
        question=question,
        slug="something",
        created_by_user="asdf",
        created_by_group="foobar",
    ).exists()



from collections import defaultdict
from contextlib import contextmanager
from functools import partial

from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context

from ..caluma_core.jexl import JEXL, ExtractTransformSubjectAnalyzer
from .models import Question


class QuestionMissing(Exception):
    pass


class QuestionValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "answer" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid question slug."

        yield from super().visit_Transform(transform)


class QuestionJexl(JEXL):
    def __init__(self, validation_context=None, **kwargs):

        if validation_context:
            if "jexl_cache" not in validation_context:
                validation_context["jexl_cache"] = defaultdict(dict)
            self._cache = validation_context["jexl_cache"]
        else:
            self._cache = defaultdict(dict)

        super().__init__(**kwargs)

        self._structure = None
        self._form = None

        context_data = None

        if validation_context:
            # cleaned up variant
            self._form = validation_context.get("form")
            self._structure = validation_context.get("structure")
            context_data = {
                "form": self._form.slug if self._form else None,
                "info": self._structure,
            }

        self.context = Context(context_data)

        self.add_transform("answer", self.answer_transform)
        self.add_transform(
            "mapby", lambda arr, key: [obj.get(key, None) for obj in arr]
        )
        self.add_binary_operator(
            "intersects", 20, lambda left, right: any(x in right for x in left)
        )

    def answer_transform(self, question_slug):
        question = self._question(question_slug)

        if self.is_hidden(question):
            if question.type in [
                Question.TYPE_MULTIPLE_CHOICE,
                Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
                Question.TYPE_TABLE,
            ]:
                return []

            return None

        return self._structure.get_field(question_slug).value()

    def validate(self, expression, **kwargs):
        return super().validate(expression, QuestionValidatingAnalyzer)

    def extract_referenced_questions(self, expr):
        transforms = ["answer", "mapby"]
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=transforms)
        )

    def _question(self, slug):
        field = self._structure.get_field(slug)
        if field:
            return field.question

        raise QuestionMissing(
            f"Question `{slug}` could not be found in form {self.context['form']}"
        )

    @contextmanager
    def use_question_context(self, question_slug):
        """Context manger to temporarily overwrite self._structure.

        This is used so we can evaluate each JEXL expression in the context
        of the corresponding question, not from where the question was
        referenced.
        This is relevant in table questions and form questions, so we always
        lookup the correct answer value (no "crosstalk" between rows, for example)
        """

        # field's parent is the fieldset - which is a valid structure object
        old_structure = self._structure
        self._structure = self._structure.get_field(question_slug).parent()
        yield
        self._structure = old_structure

    def _all_containers_hidden(self, question):
        """Check whether all containers of the given question are hidden.

        The question could be used in multiple sub-forms in the given
        document structure. This goes through all paths from the main form
        to the given question, and checks if they're hidden.

        If all subforms are hidden where the question shows up,
        the question shall be hidden as well.
        """
        paths = self._structure.root().paths_to_question(question.slug)

        res = bool(paths) and all(
            # the "inner" check here represents a single path. If any
            # question is hidden, the question is not reachable via
            # this path
            bool(path) and any(self.is_hidden(fq) for fq in path if fq != question)
            for path in paths
        )
        # The outer check verifies if all paths are "hidden" (ie have a hidden
        # question in them). If all paths from the root form to the
        # question are hidden, the question must indeed be hidden itself.

        return res

    def _cache_key(self, question_slug):
        field = self._structure.get_field(question_slug)
        return (field.document.pk, question_slug)

    def is_hidden(self, question):
        """Return True if the given question is hidden.

        This checks whether the dependency questions are hidden, then
        evaluates the question's is_hidden expression itself.
        """
        cache_key = self._cache_key(question.pk)

        if cache_key in self._cache["hidden"]:
            return self._cache["hidden"][cache_key]
        # Check visibility of dependencies before actually evaluating the `is_hidden`
        # expression. If all dependencies are hidden,
        # there is no way to evaluate our own visibility, so we default to
        # hidden state as well.

        deps = list(self.extract_referenced_questions(question.is_hidden))

        # all() returns True for the empty set, thus we need to
        # check that we have some deps at all first
        all_deps_hidden = bool(deps) and all(
            self.is_hidden(self._question(dep)) for dep in deps
        )
        if all_deps_hidden:
            self._cache["hidden"][cache_key] = True
            return True

        # Also check if the question is hidden indirectly,
        # for example via parent formquestion.
        if self._all_containers_hidden(question):
            # no way this is shown somewhere
            self._cache["hidden"][cache_key] = True
            return True
        # if the question is visible-in-context and not hidden by invisible dependencies,
        # we can evaluate it's own is_hidden expression

        with self.use_question_context(question.pk):
            self._cache["hidden"][cache_key] = self.evaluate(question.is_hidden)
        return self._cache["hidden"][cache_key]

    def is_required(self, question_field):
        cache_key = (question_field.document.pk, question_field.question.pk)
        question = question_field.question

        if cache_key in self._cache["required"]:
            return self._cache["required"][cache_key]

        deps = list(self.extract_referenced_questions(question.is_required))

        if bool(deps) and all(self.is_hidden(self._question(dep)) for dep in deps):
            # all dependent questions are hidden. cannot evaluate,
            # so assume requiredness to be False
            ret = False
        else:
            with self.use_question_context(question.pk):
                ret = self.evaluate(question.is_required)
        self._cache["required"][cache_key] = ret
        return ret



from caluma.caluma_core.events import on
from caluma.caluma_workflow.events import completed_case, completed_work_item

from .. import models


def test_events(db, work_item_factory, schema_executor):
    work_item = work_item_factory(status=models.WorkItem.STATUS_READY, child_case=None)

    @on(completed_work_item)
    def complete_work_item_event_receiver(sender, work_item, **kwargs):
        work_item.meta = {"been-there": "done that"}
        work_item.save()
        raise Exception("this should not cause an error")

    @on(completed_case)
    def complete_case_event_receiver(sender, case, **kwargs):
        case.meta = {"been-there": "done that"}
        case.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    work_item.refresh_from_db()
    assert work_item.meta == {"been-there": "done that"}
    assert work_item.case.meta == {"been-there": "done that"}



"""Setuptools package definition."""

import distutils.cmd
import os
from os import path

from setuptools import find_packages, setup

version = {}
with open("caluma/caluma_metadata.py") as fp:
    exec(fp.read(), version)

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


pipenv_setup = """
echo UID=$(id -u) > .env
echo ENV=dev >> .env
docker-compose up -d db
rm -f Pipfile*
touch Pipfile
pipenv install --python 3.7 -r requirements.txt
pipenv install -d -r requirements-dev.txt
""".strip().splitlines()


def deps_from_file(filename):
    lines = [line.strip().split("#")[0] for line in open(filename).readlines()]
    # filter out comment lines
    return [line for line in lines if line]


dependencies = deps_from_file("requirements.txt")
dev_dependencies = deps_from_file("requirements-dev.txt")


class PipenvCommand(distutils.cmd.Command):
    description = "setup local development with pipenv"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for cmd in pipenv_setup:
            assert os.system(cmd) == 0
        print(
            "\npipenv run pytest      runs the tests"
            "\npipenv shell           enters the virtualenv"
        )


setup(
    cmdclass={"pipenv": PipenvCommand},
    name=version["__title__"],
    version=version["__version__"],
    description=version["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://projectcaluma.github.io/",
    download_url="https://github.com/projectcaluma/caluma",
    license="License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    install_requires=dependencies,
)



import logging

from django.dispatch import receiver as on  # noqa

logger = logging.getLogger(__name__)


def send_event(signal, **kwargs):
    """
    Send events and handle exceptions in receiver functions.

    This wrapper handles the sending of signals as well as handling of any Exceptions that
    occur. Exceptions are logged with `ERROR` level, but not raised, in order to prevent
    rollback of the transaction.
    """
    result = signal.send_robust(**kwargs)
    for func, exc in result:
        if exc is not None:
            logger.error(
                f'Error in event handler "{func.__module__}.{func.__name__}":',
                exc_info=exc,
            )
    return result


class SendEventSerializerMixin:
    def send_event(self, event, **kwargs):
        send_event(event, sender=self.context["mutation"], **kwargs)  # noqa



from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.utils import timezone


def test_add_case_families(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_workflow"
    migrate_from = [(app, "0019_slugfield_length")]
    migrate_to = [(app, "0020_case_families")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Case = old_apps.get_model(app, "Case")
    Workflow = old_apps.get_model(app, "Workflow")
    WorkItem = old_apps.get_model(app, "WorkItem")
    Task = old_apps.get_model(app, "Task")
    HistoricalCase = old_apps.get_model(app, "HistoricalCase")

    workflow = Workflow.objects.create(slug="main-workflow")

    case = Case.objects.create(workflow=workflow)
    child_case = Case.objects.create(workflow=workflow)
    task = Task.objects.create()
    WorkItem.objects.create(child_case=child_case, case=case, task=task)
    WorkItem.objects.create(child_case=None, case=child_case, task=task)

    historical_case = HistoricalCase.objects.create(
        workflow=workflow,
        id=case.pk,
        created_at=timezone.now(),
        modified_at=timezone.now(),
        history_date=timezone.now(),
    )
    historical__child_case = HistoricalCase.objects.create(
        workflow=workflow,
        id=child_case.pk,
        created_at=timezone.now(),
        modified_at=timezone.now(),
        history_date=timezone.now(),
    )
    historical_orphan_case = HistoricalCase.objects.create(
        workflow=workflow,
        created_at=timezone.now(),
        modified_at=timezone.now(),
        history_date=timezone.now(),
    )

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)
    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    Case = new_apps.get_model(app, "Case")
    HistoricalCase = new_apps.get_model(app, "HistoricalCase")

    # Refresh objects
    case = Case.objects.get(pk=case.pk)
    historical_case = HistoricalCase.objects.get(pk=historical_case.pk)
    historical__child_case = HistoricalCase.objects.get(pk=historical__child_case.pk)
    historical_orphan_case = HistoricalCase.objects.get(pk=historical_orphan_case.pk)

    for c in Case.objects.iterator():
        assert c.family.pk == case.pk

    assert historical_case.family_id == historical__child_case.family_id == case.id
    assert historical_orphan_case.family_id == historical_orphan_case.id


def test_workitem_name_description(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_workflow"
    migrate_from = [(app, "0021_work_item_controlling_groups")]
    migrate_to = [(app, "0022_workitem_name_description")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Case = old_apps.get_model(app, "Case")
    Workflow = old_apps.get_model(app, "Workflow")
    WorkItem = old_apps.get_model(app, "WorkItem")
    Task = old_apps.get_model(app, "Task")
    HistoricalWorkItem = old_apps.get_model(app, "HistoricalWorkItem")

    workflow = Workflow.objects.create(slug="main-workflow")

    case = Case.objects.create(workflow=workflow)
    task = Task.objects.create(name="task_name", description="task_description")
    work_item = WorkItem.objects.create(child_case=None, case=case, task=task)

    HistoricalWorkItem.objects.create(
        id=work_item.pk,
        child_case=None,
        case=case,
        task=task,
        created_at=timezone.now(),
        modified_at=timezone.now(),
        history_date=timezone.now(),
    )

    HistoricalWorkItem.objects.create(
        id=work_item.pk,
        child_case=None,
        case=case,
        task=task,
        created_at=timezone.now() + timezone.timedelta(hours=1),
        modified_at=timezone.now() + timezone.timedelta(hours=1),
        history_date=timezone.now() + timezone.timedelta(hours=1),
    )

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)
    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    WorkItem = new_apps.get_model(app, "WorkItem")
    HistoricalWorkItem = new_apps.get_model(app, "HistoricalWorkItem")

    # Refresh objects
    work_item = WorkItem.objects.get(pk=work_item.pk)
    historical_work_items = HistoricalWorkItem.objects.filter(id=work_item.pk)

    for wi in [work_item, *historical_work_items]:
        assert wi.name == wi.task.name
        assert wi.description == wi.task.description



from logging import getLogger

log = getLogger(__name__)


def is_iterable_and_no_string(value):
    try:
        iter(value)
        return not isinstance(value, str)
    except TypeError:
        return False


def _field_name(db_field):
    if db_field.is_relation:
        return db_field.db_column or f"{db_field.name}_id"
    return db_field.db_column or db_field.name


def col_type_from_db(field, connection):
    """Return a (type, length) tuple for the given field."""

    sql_extract_type = """
        SELECT data_type, character_maximum_length
        FROM information_schema.columns
        WHERE
            column_name = %s AND
            table_name = %s
    """

    cursor = connection.cursor()
    cursor.execute(sql_extract_type, [_field_name(field), field.model._meta.db_table])
    res = cursor.fetchone()
    return res


def fix_foreign_key_types(apps, connection):
    """Ensure all foreign keys pointing to slug fields are correct type."""

    # I'm really not happy that this code needs to exist. I'ts a sign that
    # we somehow messed up and I couldn't reproduce exactly why this even
    # happened in the first place.  Figuring out why it happened would be
    # good, but since the broken version is out there anyway, we need the
    # fix no matter what.

    fix_sql = """
        ALTER TABLE %s
        ALTER COLUMN %s TYPE %s
    """

    for models in apps.all_models.values():
        for model in models.values():
            slug_fks = [
                field
                for field in model._meta.fields
                if field.is_relation and field.target_field.name == "slug"
            ]

            for field in slug_fks:
                # ok we have a slug foreign key
                fk_params = field.db_parameters(connection)
                target_params = field.target_field.db_parameters(connection)

                # verify django-internal specified type
                assert (
                    fk_params["type"] == target_params["type"]
                ), f"Foreign key field {field}: type mismatch with destination in django-internal representation"

                # check if the DB agrees
                fk_dbtype = col_type_from_db(field, connection)
                target_dbtype = col_type_from_db(field.target_field, connection)
                if fk_dbtype != target_dbtype:
                    cursor = connection.cursor()
                    type_, length = target_dbtype
                    new_type = f"{type_}({length})"
                    log.warning("Correcting data type of {field} to {new_type}")
                    # not using the parametrized syntax here (sql, params) as
                    # we do NOT want escaping here
                    cursor.execute(
                        fix_sql % (model._meta.db_table, _field_name(field), new_type)
                    )



from simple_history.utils import bulk_create_with_history

from ..caluma_form.models import Document
from . import models
from .jexl import GroupJexl


def get_group_jexl_structure(work_item_created_by_group, case, prev_work_item=None):
    return {
        "case": {"created_by_group": case.created_by_group},
        "work_item": {"created_by_group": work_item_created_by_group},
        "prev_work_item": {
            "controlling_groups": prev_work_item.controlling_groups
            if prev_work_item
            else None
        },
    }


def get_jexl_groups(jexl, task, case, work_item_created_by_user, prev_work_item=None):
    if jexl:
        return GroupJexl(
            validation_context=get_group_jexl_structure(
                work_item_created_by_user.group, case, prev_work_item
            ),
            task=task,
            case=case,
            work_item_created_by_user=work_item_created_by_user,
            prev_work_item=prev_work_item,
        ).evaluate(jexl)

    return []


def bulk_create_work_items(tasks, case, user, prev_work_item=None):
    work_items = []
    for task in tasks:
        controlling_groups = get_jexl_groups(
            task.control_groups,
            task,
            case,
            user,
            prev_work_item if prev_work_item else None,
        )
        addressed_groups = [
            get_jexl_groups(
                task.address_groups,
                task,
                case,
                user,
                prev_work_item if prev_work_item else None,
            )
        ]
        if task.is_multiple_instance:
            addressed_groups = [[x] for x in addressed_groups[0]]

        for groups in addressed_groups:
            work_items.append(
                models.WorkItem(
                    addressed_groups=groups,
                    controlling_groups=controlling_groups,
                    task_id=task.pk,
                    deadline=task.calculate_deadline(),
                    document=Document.objects.create_document_for_task(task, user),
                    case=case,
                    status=models.WorkItem.STATUS_READY,
                    created_by_user=user.username,
                    created_by_group=user.group,
                )
            )

    bulk_create_with_history(work_items, models.WorkItem)
    return work_items
